#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os
import shutil


def preprocess(i):

    os_info = i['os_info']

    if os_info['platform'] == 'windows':
        return {'return': 1, 'error': 'Windows is not supported in this script yet'}
    env = i['env']

    if env.get('CM_MLPERF_SKIP_RUN', '') == "yes":
        return {'return': 0}

    import json
    if 'CM_MODEL' not in env:
        return {
            'return': 1, 'error': 'Please select a variation specifying the model to run'}
    if 'CM_MLPERF_BACKEND' not in env:
        return {'return': 1,
                'error': 'Please select a variation specifying the backend'}
    if 'CM_MLPERF_DEVICE' not in env:
        return {
            'return': 1, 'error': 'Please select a variation specifying the device to run on'}

    ml_model = env['CM_MODEL']
    master_model = ml_model.replace("-99.9", "").replace("-99", "")
    master_model = master_model.replace("gptj", "gpt-j")

    backend = env['CM_MLPERF_BACKEND']
    device = env['CM_MLPERF_DEVICE']
    code_base_folder = backend + '-' + device
    if env.get('CM_MLPERF_INFERENCE_CODE_VERSION', '') == 'v4.0':
        if 'gptj' in ml_model:
            code_base_folder = "ITREX"
    if 'dlrm-v2' in ml_model:
        code_base_folder = "pytorch-cpu-int8"

    harness_root = os.path.join(
        env['CM_MLPERF_INFERENCE_RESULTS_PATH'],
        'closed',
        'Intel',
        'code',
        ml_model,
        code_base_folder)

    env['CM_HARNESS_CODE_ROOT'] = harness_root

    if env.get('CM_MODEL') == "resnet50":
        pass

    elif "bert" in env.get('CM_MODEL'):
        pass
    elif "retinanet" in env.get('CM_MODEL'):
        pass
    elif "gptj" in env.get('CM_MODEL'):
        env['CHECKPOINT_DIR'] = env['GPTJ_CHECKPOINT_PATH']

    script_path = i['run_script_input']['path']
    if env['CM_MODEL'] == "retinanet":
        env['CM_DATASET_LIST'] = env['CM_DATASET_ANNOTATIONS_FILE_PATH']

    if 'CM_MLPERF_CONF' not in env:
        env['CM_MLPERF_CONF'] = os.path.join(
            env['CM_MLPERF_INFERENCE_SOURCE'], "mlperf.conf")
    if 'CM_MLPERF_USER_CONF' not in env:
        env['CM_MLPERF_USER_CONF'] = os.path.join(
            env['CM_MLPERF_INFERENCE_CLASSIFICATION_AND_DETECTION_PATH'], "user.conf")

    loadgen_mode = env['CM_MLPERF_LOADGEN_MODE']
    env['CONDA_PREFIX'] = env['CM_CONDA_PREFIX']

    if env['CM_LOCAL_MLPERF_INFERENCE_INTEL_RUN_MODE'] == "calibration":
        if master_model == "resnet50":
            i['run_script_input']['script_name'] = "prepare_imagenet_calibration"
        elif master_model == "3d-unet":
            i['run_script_input']['script_name'] = "prepare_3d-unet_data_model"
        elif "dlrm-v2" in master_model:
            i['run_script_input']['script_name'] = "calibrate_dlrm_v2_model"
        else:
            calibration_root = os.path.join(
                env['CM_MLPERF_INFERENCE_RESULTS_PATH'],
                'closed',
                'Intel',
                'calibration',
                master_model,
                backend + "-" + device)

            if "gpt" in env['CM_MODEL']:
                i['run_script_input']['script_name'] = "calibrate_gptj_int4_model"
                calibration_path = os.path.join(calibration_root, "INT4")
                env['CM_MLPERF_INFERENCE_INTEL_CALIBRATION_PATH'] = calibration_path
                env['INT4_CALIBRATION_DIR'] = os.path.join(
                    calibration_path, "data", "quantized-int4-model")

    elif env['CM_LOCAL_MLPERF_INFERENCE_INTEL_RUN_MODE'] == "compilation":
        if master_model == "resnet50":
            i['run_script_input']['script_name'] = "compile_resnet50"
        elif master_model == "retinanet":
            i['run_script_input']['script_name'] = "compile_retinanet"
            env['CM_ML_MODEL_RETINANET_INT8_FILE_WITH_PATH'] = os.path.join(
                os.path.dirname(env['CM_ML_MODEL_FILE_WITH_PATH']), 'retinanet-int8-model.pth')

    elif env['CM_LOCAL_MLPERF_INFERENCE_INTEL_RUN_MODE'] == "build_harness":
        print(f"Harness Root: {harness_root}")
        if "bert" in env['CM_MODEL']:
            i['run_script_input']['script_name'] = "build_bert_harness"
            env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH'] = os.path.join(
                os.getcwd(), "harness", "build", "bert_inference")
            env['DATA_PATH'] = os.path.join(os.getcwd(), "harness", "bert")
        elif "stable-diffusion" in env['CM_MODEL']:
            i['run_script_input']['script_name'] = "build_sdxl_harness"
        elif "resnet50" in env['CM_MODEL']:
            i['run_script_input']['script_name'] = "build_resnet50_harness"
            env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH'] = os.path.join(
                os.getcwd(), "harness", "build", "resnet50_inference")
            env['DATA_PATH'] = os.path.join(os.getcwd(), "harness", "resnet50")
        elif "retinanet" in env['CM_MODEL']:
            i['run_script_input']['script_name'] = "build_retinanet_harness"
            env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH'] = os.path.join(
                os.getcwd(), "harness", "build", "retinanet_inference")
        elif "gpt" in env['CM_MODEL']:
            i['run_script_input']['script_name'] = "build_gptj_harness"
            env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH'] = os.path.join(
                os.getcwd(), "harness", "build", "gptj_inference")
            env['DATA_PATH'] = os.path.join(os.getcwd(), "harness", "gptj")
            env['MLPERF_INFERENCE_ROOT'] = env['CM_MLPERF_INFERENCE_SOURCE']
            if env.get('INTEL_GPTJ_INT4', '') == 'yes':
                model_precision = "int4"
                if env.get('CM_MLPERF_INFERENCE_CODE_VERSION', '') == 'v3.1':
                    env['RUN_QUANTIZATION_CMD'] = "bash run_quantization_int4.sh"
                else:
                    env['FILE_TAG'] = "final"
                    env['OUT_DIR'] = os.getcwd()
                    env['RUN_QUANTIZATION_CMD'] = "bash run_quantization.sh"
            else:
                model_precision = "int8"
                env['RUN_QUANTIZATION_CMD'] = "bash run_quantization.sh"
            if env.get('CM_MLPERF_INFERENCE_CODE_VERSION', '') == "v3.1":
                final_model_path = os.path.join(
                    harness_root, "data", f"gpt-j-{model_precision}-model", "best_model.pt")
            else:
                final_model_path = os.path.join(
                    env['OUT_DIR'], "checkpoint-final-final-q4-j-int8-pc.bin")
            model_dir_name = f"{model_precision.upper()}_MODEL_DIR"
            env[model_dir_name] = os.path.dirname(final_model_path)
            if not os.path.exists(env[model_dir_name]):
                os.makedirs(env[model_dir_name])
            env['CM_ML_MODEL_PATH'] = env[model_dir_name]
            env['CM_ML_MODEL_FILE_WITH_PATH'] = final_model_path
            if env.get('CM_MLPERF_INFERENCE_INTEL_GPTJ_INT8_MODEL_PATH',
                       '') != '' and env.get('INT8_MODEL_DIR', '') != '':
                shutil.copy(
                    env['CM_MLPERF_INFERENCE_INTEL_GPTJ_INT8_MODEL_PATH'],
                    env[model_dir_name])
            if env.get('CM_MLPERF_INFERENCE_INTEL_GPTJ_INT4_MODEL_PATH',
                       '') != '' and env.get('INT4_MODEL_DIR', '') != '':
                shutil.copy(
                    env['CM_MLPERF_INFERENCE_INTEL_GPTJ_INT4_MODEL_PATH'],
                    env[model_dir_name])

    elif env['CM_LOCAL_MLPERF_INFERENCE_INTEL_RUN_MODE'] == "run_harness":
        print(f"Harness Root: {harness_root}")
        if env.get('CM_MLPERF_LOADGEN_MODE', '') == "compliance":
            audit_path = env['CM_MLPERF_INFERENCE_AUDIT_PATH']
            shutil.copy(audit_path, env['CM_RUN_DIR'])

        if env['CM_MLPERF_LOADGEN_MODE'] == "accuracy":
            env['LOADGEN_MODE'] = 'Accuracy'
        else:
            env['LOADGEN_MODE'] = 'Performance'

        if 'bert' in env['CM_MODEL']:
            env['MODEL_PATH'] = os.path.dirname(os.path.dirname(
                env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH']))
            env['DATASET_PATH'] = os.path.dirname(os.path.dirname(
                env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH']))
            env['CM_RUN_DIR'] = i['run_script_input']['path']
            env['CM_RUN_CMD'] = "bash run_bert_harness.sh " + \
                ("--accuracy" if env['CM_MLPERF_LOADGEN_MODE']
                 == "accuracy" else "")

        elif 'resnet50' in env['CM_MODEL']:
            env['MODEL_PATH'] = os.path.dirname(os.path.dirname(
                env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH']))
            env['DATASET_PATH'] = os.path.dirname(os.path.dirname(
                env['CM_MLPERF_INFERENCE_INTEL_HARNESS_PATH']))
            env['CM_RUN_DIR'] = env['CM_MLPERF_OUTPUT_DIR']
            env['CM_RUN_CMD'] = f"bash {os.path.join(i['run_script_input']['path'],'run_resnet50_harness.sh')} "

        elif 'retinanet' in env['CM_MODEL']:
            env['MODEL_PATH'] = env['CM_ML_MODEL_RETINANET_INT8_FILE_WITH_PATH']
            env['DATA_DIR'] = env['CM_DATASET_PATH_ROOT']
            env['CM_RUN_DIR'] = env['CM_MLPERF_OUTPUT_DIR']
            env['CM_RUN_CMD'] = f"bash {os.path.join(i['run_script_input']['path'],'run_retinanet_harness.sh')} "

        elif '3d-unet' in env['CM_MODEL']:
            env['CM_RUN_DIR'] = env['CM_MLPERF_OUTPUT_DIR']
            env['CM_RUN_CMD'] = f"bash {os.path.join(i['run_script_input']['path'],'run_3d-unet_harness.sh')} "

        elif 'dlrm' in env['CM_MODEL']:
            env['CM_RUN_DIR'] = i['run_script_input']['path']
            env['CM_RUN_CMD'] = f"bash {os.path.join(i['run_script_input']['path'],'run_dlrm_v2_harness.sh')} "

        elif 'stable-diffusion' in env['CM_MODEL']:
            env['CM_RUN_DIR'] = i['run_script_input']['path']
            env['CM_RUN_CMD'] = "bash run_sdxl_harness.sh " + \
                ("--accuracy" if env['CM_MLPERF_LOADGEN_MODE']
                 == "accuracy" else "")

        elif "gptj" in env['CM_MODEL']:
            env['CM_RUN_DIR'] = i['run_script_input']['path']
            if env.get('CM_MLPERF_INFERENCE_CODE_VERSION', '') == "v3.1":
                env['CM_RUN_CMD'] = "bash run_gptj_harness_v3_1.sh "
                if env.get('INTEL_GPTJ_INT4', '') == 'yes':
                    model_precision = "int4"
                    env['INT4_MODEL_DIR'] = env['CM_ML_MODEL_PATH']
                    env['QUANTIZED_MODEL'] = os.path.join(
                        env['INT4_MODEL_DIR'], "best_int4_model.pt")
                    env['PRECISION'] = "int4_bf16_mixed"
                else:
                    env['INT8_MODEL_DIR'] = env['CM_ML_MODEL_PATH']
                    env['QUANTIZED_MODEL'] = os.path.join(
                        env["INT8_MODEL_DIR"], "best_model.pt")
                    env['PRECISION'] = "int8"
            elif env.get('CM_MLPERF_INFERENCE_CODE_VERSION', '') == "v4.0":
                env['CM_RUN_CMD'] = "bash run_gptj_harness_v4_0.sh "

                if env['CM_MLPERF_RUN_STYLE'] == "test":
                    env['TOTAL_SAMPLE_COUNT'] = env['CM_TEST_QUERY_COUNT']
                else:
                    env['TOTAL_SAMPLE_COUNT'] = env.get(
                        'CM_MLPERF_MAX_QUERY_COUNT', env['CM_TEST_QUERY_COUNT'])

                if env['CM_MLPERF_LOADGEN_SCENARIO'] == "Offline":
                    env['WORKERS_PER_PROC'] = 4
                else:
                    env['WORKERS_PER_PROC'] = 1

    return {'return': 0}


def postprocess(i):

    env = i['env']
    state = i['state']

    return {'return': 0}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os
import sys
import yaml


def preprocess(i):

    os_info = i['os_info']

    env = i['env']

    meta = i['meta']

    automation = i['automation']

    quiet = (env.get('CM_QUIET', False) == 'yes')

    if env.get('CM_CREATE_INPUT_BATCH', '') == 'yes':
        r = create_batched_inputs(env)
        if r['return'] > 0:
            return r

    r = construct_calibration_cmd(env)
    if r['return'] > 0:
        return r
    cmd = r['cmd']

    print("Profiling from " + os.getcwd())

    env['CM_RUN_CMD'] = cmd

    return {'return': 0}


def create_batched_inputs(env):
    original_images_file = env['CM_DATASET_PREPROCESSED_IMAGES_LIST']
    batchsize = env['CM_QAIC_MODEL_BATCH_SIZE']

    file_paths = []
    with open(original_images_file) as f:
        file_paths = f.read().splitlines()

    i = 0
    outfile = None
    lastfile = None
    outfiles = []
    os.makedirs(os.path.join(os.getcwd(), "raw"), exist_ok=True)
    for file in file_paths:
        if i % int(batchsize) == 0:
            filename = os.path.basename(file).replace(".rgb32", ".raw")
            outfile = os.path.join(os.getcwd(), "raw", filename)
            outfiles.append(outfile)
            with open(outfile, "wb") as f:
                pass
        with open(outfile, "ab") as f:
            with open(file, "rb") as infile:
                f.write(infile.read())
        i = i + 1
        lastfile = file

    while i % int(batchsize) != 0:
        with open(outfile, "ab") as f:
            with open(lastfile, "rb") as infile:
                f.write(infile.read())
        i = i + 1
    with open("batched_input_files", "w") as f:
        f.write("\n".join(outfiles))

    return {'return': 0}


def construct_calibration_cmd(env):
    compiler_params = env['CM_QAIC_COMPILER_PARAMS']
    batchsize = env.get('CM_QAIC_MODEL_BATCH_SIZE', "1")
    cmd = env['CM_QAIC_EXEC_PATH'] + " "
    if env.get('CM_CREATE_INPUT_BATCH', '') == 'yes':
        cmd += " -input-list-file=batched_input_files  -batchsize=" + batchsize + " "
    cmd += compiler_params + " -dump-profile=profile.yaml -model=" + \
        env['CM_ML_MODEL_FILE_WITH_PATH']

    return {'return': 0, 'cmd': cmd}


def postprocess(i):

    env = i['env']
    profile_file_path = os.path.join(os.getcwd(), "profile.yaml")
    env['CM_QAIC_MODEL_PROFILE_WITH_PATH'] = profile_file_path

    if env.get('CM_ML_MODEL_INPUT_LAYER_NAME', '') != '':
        input_layer_names = [env.get('CM_ML_MODEL_INPUT_LAYER_NAME')]
    else:
        input_layer_names = ["images:0", "images/:0"]

    output_layer_names_conf = [[], []]
    output_layer_names_loc = [[], []]

    output_layer_names_loc[0] = [
        "/GatherElements/:0",
        "/GatherElements_1/:0",
        "/GatherElements_2/:0",
        "/GatherElements_3/:0",
        "/GatherElements_4/:0"
    ]

    output_layer_names_conf[0] = [
        "/TopK/:0",
        "/TopK_1/:0",
        "/TopK_2/:0",
        "/TopK_3/:0",
        "/TopK_4/:0"
    ]

    output_layer_names_loc[1] = [
        "GatherElements_588/:0",
        "GatherElements_598/:0",
        "GatherElements_608/:0",
        "GatherElements_618/:0",
        "GatherElements_628/:0"
    ]

    output_layer_names_conf[1] = [
        "TopK_570/:0",
            "TopK_572/:0",
            "TopK_574/:0",
            "TopK_576/:0",
            "TopK_578/:0"
        ]

    if env.get('CM_QAIC_MODEL_NAME', '') == "retinanet":
        with open(profile_file_path, "r") as stream:
            try:
                output_min_val_loc = sys.maxsize
                output_max_val_loc = -sys.maxsize
                output_min_val_conf = sys.maxsize
                output_max_val_conf = -sys.maxsize
                docs = yaml.load_all(stream, yaml.FullLoader)
                for doc in docs:


if isinstance(doc,                     if )
                        node_names = [k['NodeOutputName'] for k in doc]
                        oindex = None

                        for output in output_layer_names_loc:
                            if output[0] in node_names:
                                oindex = output_layer_names_loc.index(output)
                                break

                        if oindex is None:
                            return {
                                'return': 1, 'error': 'Output node names not found for the given retinanet model'}

                        for k in doc:
                            if k["NodeOutputName"] in input_layer_names:
                                min_val = k['Min']
                                max_val = k['Max']
                                scale, offset = get_scale_offset(
                                    min_val, max_val)
                                env['CM_QAIC_MODEL_RETINANET_IMAGE_SCALE'] = scale
                                env['CM_QAIC_MODEL_RETINANET_IMAGE_OFFSET'] = offset

                            if k["NodeOutputName"] in output_layer_names_loc[oindex]:
                                min_val = k['Min']
                                max_val = k['Max']
                                if min_val < output_min_val_loc:
                                    output_min_val_loc = min_val
                                if max_val > output_max_val_loc:
                                    output_max_val_loc = max_val
                                loc_scale, loc_offset = get_scale_offset(
                                    min_val, max_val)
                                index = output_layer_names_loc[oindex].index(
                                    k["NodeOutputName"])
                                env[f'CM_QAIC_MODEL_RETINANET_LOC_SCALE{index}'] = loc_scale
                                # to uint8 is done in NMS code
                                env[f'CM_QAIC_MODEL_RETINANET_LOC_OFFSET{index}'] = loc_offset - 128

                                total_range = max_val - min_val
                                scale = total_range / 256.0
                                offset = round(-min_val / scale)

                            if k["NodeOutputName"] in output_layer_names_conf[oindex]:
                                min_val = k['Min']
                                max_val = k['Max']
                                if min_val < output_min_val_conf:
                                    output_min_val_conf = min_val
                                if max_val > output_max_val_conf:
                                    output_max_val_conf = max_val
                                conf_scale, conf_offset = get_scale_offset(
                                    min_val, max_val)
                                index = output_layer_names_conf[oindex].index(
                                    k["NodeOutputName"])
                                env[f'CM_QAIC_MODEL_RETINANET_CONF_SCALE{index}'] = conf_scale
                                # to uint8 is done in NMS code
                                env[f'CM_QAIC_MODEL_RETINANET_CONF_OFFSET{index}'] = conf_offset - 128
                                total_range = max_val - min_val
                                scale = total_range / 256.0
                                offset = round(-min_val / scale)

                loc_scale, loc_offset = get_scale_offset(
                    output_min_val_loc, output_max_val_loc)
                conf_scale, conf_offset = get_scale_offset(
                    output_min_val_conf, output_max_val_conf)
                env['CM_QAIC_MODEL_RETINANET_LOC_SCALE'] = loc_scale
                env['CM_QAIC_MODEL_RETINANET_LOC_OFFSET'] = loc_offset - 128  # to uint8 is done in NMS code
                env['CM_QAIC_MODEL_RETINANET_CONF_SCALE'] = conf_scale
                env['CM_QAIC_MODEL_RETINANET_CONF_OFFSET'] = conf_offset - 128  # to uint8 is done in NMS code

            except yaml.YAMLError as exc:
                return {'return': 1, 'error': exc}

    return {'return': 0}


def get_scale_offset(min_val, max_val):
    total_range = max_val - min_val
    scale = total_range /256.0
    offset = round(-min_val / scale)
    return scale, offset



"""
abstract backend class
"""


# pylint: disable=unused-argument,missing-docstring

class Backend():
    def __init__(self):
        self.inputs = []
        self.outputs = []

    def version(self):
        raise NotImplementedError("Backend:version")

    def name(self):
        raise NotImplementedError("Backend:name")

    def load(self, model_path, inputs=None, outputs=None):
        raise NotImplementedError("Backend:load")

    def predict(self, feed):
        raise NotImplementedError("Backend:predict")



# Author: Grigori Fursin
# Contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

import os

from cmind.automation import Automation
from cmind import utils


class CAutomation(Automation):
    """
    Automation actions
    """

    ############################################################
    def __init__(self, cmind, automation_file):
        super().__init__(cmind, __file__)

    ############################################################
    def test(self, i):
        """
        Test automation

        Args:
          (CM input dict):

          (out) (str): if 'con', output to console

          automation (str): automation as CM string object

          parsed_automation (list): prepared in CM CLI or CM access function
                                    [ (automation alias, automation UID) ] or
                                    [ (automation alias, automation UID), (automation repo alias, automation repo UID) ]

          (artifact) (str): artifact as CM string object

          (parsed_artifact) (list): prepared in CM CLI or CM access function
                                    [ (artifact alias, artifact UID) ] or
                                    [ (artifact alias, artifact UID), (artifact repo alias, artifact repo UID) ]

          ...

        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

          * Output from this automation action

        """

        import json
        print(json.dumps(i, indent=2))

        return {'return': 0}



# Author: Grigori Fursin
# Contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

import os
import cmind
import copy

base_path = {}
base_path_meta = {}

##########################################################################


def load_cfg(i):

    tags = i.get('tags', '')
    artifact = i.get('artifact', '')

    key = i.get('key', '')
    key_end = i.get('key_end', [])

    ii = {'action': 'find',
        'automation': 'cfg'}
    if artifact != '':
        ii['artifact'] = artifact
    elif tags != '':
        ii['tags'] = tags

    r = cmind.access(ii)
    if r['return'] > 0:
        return r

    lst = r['list']

    prune = i.get('prune', {})
    prune_key = prune.get('key', '')
    prune_key_uid = prune.get('key_uid', '')
    prune_meta_key = prune.get('meta_key', '')
    prune_meta_key_uid = prune.get('meta_key_uid', '')
    prune_uid = prune.get('uid', '')
    prune_list = prune.get('list', [])

    # Checking individual files inside CM entry
    selection = []

    if i.get('skip_files', False):
        for l in lst:
            meta = l.meta
            full_path = l.path

            meta['full_path'] = full_path

            add = True

            if prune_key != '' and prune_key_uid != '':
                if prune_key_uid not in meta.get(prune_key, []):
                    add = False

            if add:
                selection.append(meta)
    else:
        for l in lst:
            path = l.path

            main_meta = l.meta

            skip = False

            if prune_meta_key != '' and prune_meta_key_uid != '':
                if prune_meta_key_uid not in main_meta.get(prune_meta_key, []):
                    skip = True

            if skip:
                continue

            all_tags = main_meta.get('tags', [])

            files = os.listdir(path)

            for f in files:
                if key != '' and not f.startswith(key):
                    continue

                if f.startswith('_') or (not f.endswith(
                    '.json') and not f.endswith('.yaml')):
                    continue

                if len(key_end) > 0:
                    skip = True
                    for ke in key_end:
                        if f.endswith(ke):
                            skip = False
                            break
                    if skip:
                        continue

                full_path = os.path.join(path, f)

                full_path_without_ext = full_path[:-5]

                r = cmind.utils.load_yaml_and_json(full_path_without_ext)
                if r['return'] > 0:
                    print('Warning: problem loading file {}'.format(full_path))
                else:
                    meta = r['meta']

                    # Check base
                    r = process_base(meta, full_path)
                    if r['return'] > 0:
                        return r
                    meta = r['meta']

                    uid = meta['uid']

                    # Check pruning
                    add = True

                    if len(prune) > 0:
                        if prune_uid != '' and uid != prune_uid:
                            add = False

                        if add and len(
                            prune_list) > 0 and uid not in prune_list:
                            add = False

                        if add and prune_key != '' and prune_key_uid != '' and prune_key_uid != meta.get(
                            prune_key, None):
                            add = False

                    if add:
                        meta['full_path'] = full_path

                        add_all_tags = copy.deepcopy(all_tags)

                        name = meta.get('name', '')
                        if name == '':
                            name = ' '.join(meta.get('tags', []))
                        name = name.strip()
                        meta['name'] = name

                        file_tags = meta.get('tags', '').strip()
                        if file_tags == '':
                            if name != '':
                                add_all_tags += [v.lower()
                                                         for v in name.split(' ')]
                        else:
                            add_all_tags += file_tags.split(',')

                        meta['all_tags'] = add_all_tags

                        meta['main_meta'] = main_meta

                        selection.append(meta)

    return {'return': 0, 'lst': lst, 'selection': selection}

##########################################################################


def process_base(meta, full_path):

    global base_path, base_path_meta

    _base = meta.get('_base', '')
    if _base != '':
        name = ''

        filename = _base
        full_path_base = os.path.dirname(full_path)

        if not filename.endswith('.yaml') and not filename.endswith('.json'):
            return {'return': 1, 'error': '_base file {} in {} must be .yaml or .json'.format(
                filename, full_path)}

        if ':' in _base:
            x = _base.split(':')
            name = x[0]

            full_path_base = base_path.get(name, '')
            if full_path_base == '':

                # Find artifact
                r = cmind.access({'action': 'find',
                                  'automation': 'cfg',
                                  'artifact': name})
                if r['return'] > 0:
                    return r

                lst = r['list']

                if len(lst) == 0:
                    if not os.path.isfile(path):
                        return {'return': 1, 'error': '_base artifact {} not found in {}'.format(
                            name, full_path)}

                full_path_base = lst[0].path

                base_path[name] = full_path_base

            filename = x[1]

        # Load base
        path = os.path.join(full_path_base, filename)

        if not os.path.isfile(path):
            return {'return': 1, 'error': '_base file {} not found in {}'.format(
                filename, full_path)}

        if path in base_path_meta:
            base = copy.deepcopy(base_path_meta[path])
        else:
            path_without_ext = path[:-5]

            r = cmind.utils.load_yaml_and_json(path_without_ext)
            if r['return'] > 0:
                return r

            base = r['meta']

            base_path_meta[path] = copy.deepcopy(base)

        for k in meta:
            v = meta[k]

            if k not in base:
                base[k] = v
            else:
                if isinstance(v, str):
                    # Only merge a few special keys and overwrite the rest
                    if k in ['tags', 'name']:
                        base[k] += meta[k]
                    else:
                        base[k] = meta[k]


elif isinstance(v,                 elif)                    for vv in v:
                        base[k].append(vv)
elif isinstance(v,                 elif )                    base[k].merge(v)

        meta = base

    return {'return': 0, 'meta':meta}

##########################################################################

def select_cfg(i):

    self_module = i['self_module']
    tags = i['tags']
    alias = i.get('alias', '')
    uid = i.get('uid', '')
    title = i.get('title', '')

    # Check if alias is not provided
    r = self_module.cmind.access({'action': 'find', 'automation':'cfg', 'tags':'basic,docker,configurations'})
    if r['return'] > 0:
        return r

    lst = r['list']

    selector = []

    # Do coarse-grain search for CM artifacts
    for l in lst:
        p = l.path

        if alias != '':
            for ext in ['.json', '.yaml']:
                p1 = os.path.join(p, alias +ext)
                if os.path.isfile(p1):
                    selector.append({'path': p1, 'alias':alias})
                    break

        else:
            files = os.listdir(p)

            for f in files:
                if not f.startswith('_cm') and (
                    f.endswith('.json') or f.endswith('.yaml')):
                    selector.append({'path': os.path.join(p, f), 'alias':f[:-5]})

    # Load meta for name and UID
    selector_with_meta = []
    for s in range(0, len(selector)):
        ss = selector[s]

        path = ss['path']

        full_path_without_ext = path[:-5]

        r = cmind.utils.load_yaml_and_json(full_path_without_ext)
        if r['return'] >0:
            print('Warning: problem loading configuration file {}'.format(path))

        meta = r['meta']

        if uid == '' or meta.get('uid', '') == uid:
            ss['meta'] = meta
            selector_with_meta.append(ss)

    # Quit if no configurations found
    if len(selector_with_meta) == 0:
        return {'return': 16, 'error':'configuration was not found'}

    select = 0
    if len(selector_with_meta) > 1:
        xtitle = ' ' + title if title != '' else ''
        print('')
        print('Available{} configurations:'.format(xtitle))

        print('')

        selector_with_meta = sorted(selector_with_meta, key = lambda x: x['meta'].get('name', ''))
        s = 0
        for ss in selector_with_meta:
            alias = ss['alias']
            uid = ss['meta'].get('uid', '')
            name = ss['meta'].get('name', '')

            x = name
            if x!='':
                x+=' '
            x += '(' + uid + ')'

            print(f'{s}) {x}'.format(s, x))

            s += 1

        print('')
        select = input('Enter configuration number of press Enter for 0: ')

        if select.strip() == '':
            select = '0'

    select = int(select)

    if select <0 or select>=len(selector):
        return {'return': 1, 'error':'selection is out of range'}

    ss = selector_with_meta[select]

    return {'return': 0, 'selection':ss}



import os
import pyperclip as pc

text = os.environ.get('CM_COPY_TO_CLIPBOARD_TEXT', '')

add_quotes = os.environ.get(
    'CM_COPY_TO_CLIPBOARD_TEXT_ADD_QUOTES', '') in [
        True, 'True', 'yes']

if add_quotes:
    text = '"' + text + '"'

pc.copy(text)



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os

lscpu_out = 'tmp-lscpu.out'


def preprocess(i):

    if os.path.isfile(lscpu_out):
        os.remove(lscpu_out)

    return {'return': 0}


def postprocess(i):

    state = i['state']

    env = i['env']

    os_info = i['os_info']

    automation = i['automation']
    logger = automation.cmind.logger

    if os_info['platform'] == 'windows':
        sys = []
        sys1 = []
        cpu = []
        cpu1 = []

        import csv

        try:
            f = 'tmp-systeminfo.csv'

            if not os.path.isfile(f):
                print('WARNING: {} file was not generated!'.format(f))
            else:
                keys = {}
                j = 0
                with open(f, 'r') as csvf:
                    for s in csv.reader(csvf):
                        if j == 0:
                            keys = s
                        else:
                            x = {}
                            for k in range(0, len(s)):
                                x[keys[k]] = s[k]

                            sys.append(x)

                            if j == 1:
                                sys1 = x

                        j += 1

        except Exception as e:
            logger.warning(
                'WARNING: problem processing file {} ({})!'.format(
                    f, format(e)))
            pass

        try:
            f = 'tmp-wmic-cpu.csv'
            if not os.path.isfile(f):
                logger.warning('WARNING: {} file was not generated!'.format(f))
            else:

                keys = {}
                j = 0

                with open(f, 'r', encoding='utf16') as csvf:
                    for s in csv.reader(csvf):
                        if j == 1:
                            keys = s
                        elif j > 1:
                            x = {}
                            for k in range(0, len(s)):
                                x[keys[k]] = s[k]

                            cpu.append(x)

                            if j == 2:
                                cpu1 = x

                        j += 1

        except Exception as e:
            logger.warning(
                'WARNING: problem processing file {} ({})!'.format(
                    f, format(e)))
            pass

        state['host_device_raw_info'] = {
            'sys': sys, 'sys1': sys1, 'cpu': cpu, 'cpu1': cpu1}

        logger.warning(
            'WARNING: need to unify system and cpu output on Windows')

        return {'return': 0}

    ##########################################################################
    # Linux
    if not os.path.isfile(lscpu_out):
        print('WARNING: lscpu.out file was not generated!')

        # Currently ignore this error though probably should fail?
        # But need to check that is supported on all platforms.
        return {'return': 0}

    r = utils.load_txt(file_name=lscpu_out)
    if r['return'] > 0:
        return r

    ss = r['string']

    # state['cpu_info_raw'] = ss

    # Unifying some CPU info across different platforms
    unified_env = {
        'CM_CPUINFO_CPUs': 'CM_HOST_CPU_TOTAL_CORES',
        'CM_CPUINFO_L1d_cache': 'CM_HOST_CPU_L1D_CACHE_SIZE',
        'CM_CPUINFO_L1i_cache': 'CM_HOST_CPU_L1I_CACHE_SIZE',
        'CM_CPUINFO_L2_cache': 'CM_HOST_CPU_L2_CACHE_SIZE',
        'CM_CPUINFO_L3_cache': 'CM_HOST_CPU_L3_CACHE_SIZE',
        'CM_CPUINFO_Sockets': 'CM_HOST_CPU_SOCKETS',
        'CM_CPUINFO_NUMA_nodes': 'CM_HOST_CPU_NUMA_NODES',
        'CM_CPUINFO_Cores_per_socket': 'CM_HOST_CPU_PHYSICAL_CORES_PER_SOCKET',
        'CM_CPUINFO_Cores_per_cluster': 'CM_HOST_CPU_PHYSICAL_CORES_PER_SOCKET',
        'CM_CPUINFO_Threads_per_core': 'CM_HOST_CPU_THREADS_PER_CORE',
        'CM_CPUINFO_Architecture': 'CM_HOST_CPU_ARCHITECTURE',
        'CM_CPUINFO_CPU_family': 'CM_HOST_CPU_FAMILY',
        'CM_CPUINFO_CPU_max_MHz': 'CM_HOST_CPU_MAX_MHZ',
        'CM_CPUINFO_Model_name': 'CM_HOST_CPU_MODEL_NAME',
        'CM_CPUINFO_On_line_CPUs_list': 'CM_HOST_CPU_ON_LINE_CPUS_LIST',
        'CM_CPUINFO_Vendor_ID': 'CM_HOST_CPU_VENDOR_ID',
        'CM_CPUINFO_hw_physicalcpu': 'CM_HOST_CPU_TOTAL_PHYSICAL_CORES',
        'CM_CPUINFO_hw_logicalcpu': 'CM_HOST_CPU_TOTAL_CORES',
        'CM_CPUINFO_hw_packages': 'CM_HOST_CPU_SOCKETS',
        'CM_CPUINFO_hw_memsize': 'CM_HOST_CPU_MEMSIZE',
        'CM_CPUINFO_hw_l1icachesize': 'CM_HOST_CPU_L1I_CACHE_SIZE',
        'CM_CPUINFO_hw_l1dcachesize': 'CM_HOST_CPU_L1D_CACHE_SIZE',
        'CM_CPUINFO_hw_l2cachesize': 'CM_HOST_CPU_L2_CACHE_SIZE'
    }

    if env['CM_HOST_OS_TYPE'] == 'linux':
        vkeys = ['Architecture', 'Model name', 'Vendor ID', 'CPU family', 'NUMA node(s)', 'CPU(s)',
                 'On-line CPU(s) list', 'Socket(s)', 'Core(s) per socket', 'Core(s) per cluster', 'Thread(s) per core', 'L1d cache', 'L1i cache', 'L2 cache',
                 'L3 cache', 'CPU max MHz']
    elif env['CM_HOST_OS_FLAVOR'] == 'macos':
        vkeys = ['hw.physicalcpu', 'hw.logicalcpu', 'hw.packages', 'hw.ncpu', 'hw.memsize', 'hw.l1icachesize',
                 'hw.l2cachesize']
    if vkeys:
        for s in ss.split('\n'):
            v = s.split(':')
            key = v[0]
            if key in vkeys:
                env_key = 'CM_CPUINFO_' + key.replace(
                    " ",
                    "_").replace(
                    '(',
                    '').replace(
                    ')',
                    '').replace(
                    '-',
                    '_').replace(
                    '.',
                    '_')
                if env_key in unified_env:
                    env[unified_env[env_key]] = v[1].strip()
                else:
                    env[env_key] = v[1].strip()

    if env.get('CM_HOST_CPU_SOCKETS', '') == '-':  # assume as 1
        env['CM_HOST_CPU_SOCKETS'] = '1'

    if env.get('CM_HOST_CPU_TOTAL_CORES', '') != '' and env.get(
            'CM_HOST_CPU_TOTAL_LOGICAL_CORES', '') == '':
        env['CM_HOST_CPU_TOTAL_LOGICAL_CORES'] = env['CM_HOST_CPU_TOTAL_CORES']

    if env.get('CM_HOST_CPU_TOTAL_LOGICAL_CORES', '') != '' and env.get(
            'CM_HOST_CPU_TOTAL_PHYSICAL_CORES', '') != '' and env.get('CM_HOST_CPU_THREADS_PER_CORE', '') == '':
        env['CM_HOST_CPU_THREADS_PER_CORE'] = str(int(int(env['CM_HOST_CPU_TOTAL_LOGICAL_CORES']) //
                                                      int(env['CM_HOST_CPU_TOTAL_PHYSICAL_CORES'])))

    if env.get('CM_HOST_CPU_SOCKETS', '') != '' and env.get('CM_HOST_CPU_TOTAL_PHYSICAL_CORES',
                                                            '') != '' and env.get('CM_HOST_CPU_PHYSICAL_CORES_PER_SOCKET', '') == '':
        env['CM_HOST_CPU_PHYSICAL_CORES_PER_SOCKET'] = str(
            int(env['CM_HOST_CPU_TOTAL_PHYSICAL_CORES']) // int(env['CM_HOST_CPU_SOCKETS']))

    return {'return': 0}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os
import shutil


def preprocess(i):

    os_info = i['os_info']

    if os_info['platform'] == 'windows':
        return {'return': 1, 'error': 'Windows is not supported in this script yet'}
    env = i['env']

    if env.get('CM_MLPERF_SKIP_RUN', '') == "yes":
        return {'return': 0}

    if 'CM_MODEL' not in env:
        return {
            'return': 1, 'error': 'Please select a variation specifying the model to run'}
    if 'CM_MLPERF_BACKEND' not in env:
        return {'return': 1,
                'error': 'Please select a variation specifying the backend'}
    if 'CM_MLPERF_DEVICE' not in env:
        return {
            'return': 1, 'error': 'Please select a variation specifying the device to run on'}

    r = get_run_cmd(env['CM_MODEL'], i)
    if r['return'] > 0:
        return r
    run_cmd = r['run_cmd']
    run_dir = r['run_dir']
    print(run_cmd)
    print(run_dir)
    return {'return': 1, 'error': 'Run command needs to be tested!'}


def get_run_cmd(model, i):
    env = i['env']
    if "gptj" in model:
        scenario = env['CM_MLPERF_LOADGEN_SCENARIO']
        device = env['CM_MLPERF_DEVICE']
        mode = env['CM_MLPERF_LOADGEN_MODE']
        outdir = env['CM_MLPERF_OUTPUT_DIR']
        mlperf_conf_path = env['CM_MLPERF_CONF']
        user_conf_path = env['CM_MLPERF_USER_CONF']
        api_server = env.get('CM_MLPERF_INFERENCE_API_SERVER', 'localhost')
        model_path = env['GPTJ_CHECKPOINT_PATH']
        dataset_path = env['CM_DATASET_CNNDM_EVAL_PATH']
        precision = env['CM_MLPERF_MODEL_PRECISION']
        if mode == "accuracy":
            accuracy_string = " --accuracy "
        else:
            accuracy_string = ""

        run_cmd = f"python3 -u main.py --scenario {scenario} --model-path {model_path} --api-server {api_server} --api-model-name gpt-j-cnn --mlperf-conf {mlperf_conf_path} {accuracy_string} --vllm --user-conf {user_conf_path} --dataset-path {dataset_path} --output-log-dir {outdir} --dtype float32 --device {device} "
        submitter = "CTuning"
        run_dir = os.path.join(
            env['CM_MLPERF_INFERENCE_IMPLEMENTATION_REPO'],
            "open",
            submitter,
            "code",
            "gptj-99")

        return {'return': 0, 'run_cmd': run_cmd, 'run_dir': run_dir}


def postprocess(i):

    env = i['env']

    return {'return': 0}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os
import shutil


def preprocess(i):

    os_info = i['os_info']

#    if os_info['platform'] == 'windows':
# return {'return':1, 'error': 'Windows is not supported in this script
# yet'}

    env = i['env']
    meta = i['meta']

    env_key = get_env_key(env)

    cm_git_url = env['CM_GIT_URL']

    if 'CM_GIT_REPO_NAME' not in env:
        update_env(
            env,
            'CM_GIT_REPO{}_NAME',
            env_key,
            os.path.basename(
                env['CM_GIT_URL']))

    if 'CM_GIT_DEPTH' not in env:
        env['CM_GIT_DEPTH'] = ''

    if 'CM_GIT_RECURSE_SUBMODULES' not in env:
        env['CM_GIT_RECURSE_SUBMODULES'] = ''

    if env.get('CM_GIT_CHECKOUT', '') == '':
        env['CM_GIT_CHECKOUT'] = env.get(
            'CM_GIT_SHA', env.get(
                'CM_GIT_BRANCH', ''))

    git_checkout_string = " -b " + env['CM_GIT_BRANCH'] if (
        "CM_GIT_BRANCH" in env and env.get('CM_GIT_SHA', '') == '') else ""

    git_clone_cmd = "git clone " + env['CM_GIT_RECURSE_SUBMODULES'] + git_checkout_string + " " + \
        env['CM_GIT_URL'] + " " + \
        env.get('CM_GIT_DEPTH', '') + ' ' + env['CM_GIT_CHECKOUT_FOLDER']

    env['CM_GIT_CLONE_CMD'] = git_clone_cmd
    env['CM_TMP_GIT_PATH'] = os.path.join(
        os.getcwd(), env['CM_GIT_CHECKOUT_FOLDER'], ".gitdone")

    return {'return': 0}


def postprocess(i):

    env = i['env']
    state = i['state']
    env['CM_GIT_CHECKOUT_PATH'] = os.path.join(
        os.getcwd(), env['CM_GIT_CHECKOUT_FOLDER'])
    git_checkout_path = env['CM_GIT_CHECKOUT_PATH']

    env_key = get_env_key(env)

    # We remap CM_GIT variables with CM_GIT_REPO prefix so that they don't
    # contaminate the env of the parent script
    update_env(env, 'CM_GIT_REPO{}_CHECKOUT_PATH',
               env_key, env['CM_GIT_CHECKOUT_PATH'])
    update_env(env, 'CM_GIT_REPO{}_URL', env_key, env['CM_GIT_URL'])
    update_env(env, 'CM_GIT_REPO{}_CHECKOUT', env_key, env['CM_GIT_CHECKOUT'])
    update_env(env, 'CM_GIT_REPO{}_DEPTH', env_key, env['CM_GIT_DEPTH'])
    update_env(env, 'CM_GIT_REPO{}_CHECKOUT_FOLDER',
               env_key, env['CM_GIT_CHECKOUT_FOLDER'])
    update_env(env, 'CM_GIT_REPO{}_PATCH', env_key, env['CM_GIT_PATCH'])
    update_env(env, 'CM_GIT_REPO{}_RECURSE_SUBMODULES',
               env_key, env['CM_GIT_RECURSE_SUBMODULES'])

    if (env.get('CM_GIT_CHECKOUT_PATH_ENV_NAME', '') != ''):
        env[env['CM_GIT_CHECKOUT_PATH_ENV_NAME']] = git_checkout_path

    env['CM_GET_DEPENDENT_CACHED_PATH'] = git_checkout_path

    if os.path.exists("tmp-cm-git-hash.out"):
        with open("tmp-cm-git-hash.out", "r") as f:
            git_hash = f.readline().strip()
            env['CM_GIT_REPO_CURRENT_HASH'] = git_hash

    return {'return': 0}


def get_env_key(env):

    env_key = env.get('CM_GIT_ENV_KEY', '')

    if env_key != '' and not env_key.startswith('_'):
        env_key = '_' + env_key

    return env_key


def update_env(env, key, env_key, var):

    env[key.format('')] = var

    if env_key != '':
        env[key.format(env_key)] = var

    return



# Developer(s): Grigori Fursin

import cmind
import os
import datetime
import misc


def page(st, params, extra):

    end_html = ''

    url_prefix = st.config.get_option('server.baseUrlPath') + '/'

    if not extra.get('skip_header', False):
        st.markdown('---')
        st.markdown(
            '**Install [MLCommons Collective Mind automation framework](https://github.com/mlcommons/ck):**')

    md = ''

    ###################################################################
    # Select OS
    choices = [('Ubuntu, Debian and similar Linux', 'linux'),
               ('Red Hat and CentOS', 'redhat'),
               ('MacOS', 'macos'),
               ('Windows', 'windows')]

    host_os_selection = 0

    if extra.get('run_on_windows', False):
        host_os_selection = 3

    host_os = st.selectbox('Select your host OS:',
                           range(len(choices)),
                           format_func=lambda x: choices[x][0],
                           index=host_os_selection,
                           key='install_select_host_os')

    host_os_index = choices[host_os][1]

    cur_script_file = __file__
    cur_script_path = os.path.dirname(cur_script_file)

    notes = os.path.join(cur_script_path, 'install', host_os_index + '.md')

    if os.path.isfile(notes):
        r = cmind.utils.load_txt(notes)
        if r['return'] > 0:
            return r
        s = r['string']
        if s != '':
            show = st.toggle('Show system dependencies?', value=True)
            if show:
                md += s

    need_user = ''
    python = 'python3'
    if host_os_index == 'redhat':
        need_user = ' --user'
    elif host_os_index == 'windows':
        python = 'python'

    ###################################################################
    # Select repository

    choices = [
        ('Development GitHub version: mlcommons@cm4mlops', 'dev'),
        ('Stable GitHub version: mlcommons@cm4mlops', 'main'),
        ('Stable ZIP archive from Zenodo: 20240306', 'zenodo'),
        ('Stable ZIP archive from GitHub: 20240416', 'zip-github')
    ]

    repo = st.selectbox('Select repository with [automation recipes (CM scripts)](https://access.cknowledge.org/playground/?action=scripts):',
                        range(len(choices)),
                        format_func=lambda x: choices[x][0],
                        index=0,
                        key='select_repo')

    repo_index = choices[repo][1]

    # Add stable repo from Zenodo
    if repo_index == 'dev':
        cm_repo = 'mlcommons@cm4mlops --checkout=dev'
    elif repo_index == 'zenodo':
        cm_repo = '--url=https://zenodo.org/records/10787459/files/cm-mlops-repo-20240306.zip'
    elif repo_index == 'zip-github':
        cm_repo = '--url=https://github.com/mlcommons/cm4mlops/archive/refs/tags/r20240416.zip --skip-zip-parent-dir'
    else:
        cm_repo = 'mlcommons@cm4mlops'

    x = '{} -m pip install cmind -U {}\n\n'.format(python, need_user)
    x += 'cm test core \n\n'
    x += 'cm pull repo {}\n\n'.format(cm_repo)

    clean_cm_cache = st.toggle(
        'Clean CM cache',
        value=True,
        key='install_clean_cm_cache')

    cm_clean_cache = 'cm rm cache -f\n\n' if clean_cm_cache else ''

    x += cm_clean_cache

    python_venv_name = params.get('@adr.python.name', '')
    python_ver_min = params.get('@adr.python.version_min', '')
    python_ver = params.get('@adr.python.version', '')

    if python_venv_name == '':
        use_python_venv = st.toggle(
            'Use Python Virtual Environment for CM scripts?', value=False)
        if use_python_venv:
            python_venv_name = st.text_input(
                'Enter some CM python venv name for your project:',
                value="mlperf-v4.0")

            if python_ver_min == '':
                python_ver_min = st.text_input(
                    '[Optional] Specify min version such as 3.8:')

    y = ''
    if python_venv_name != '':  # or python_ver!='' or python_ver_min!='':
        y = 'cm run script "get sys-utils-cm"\n'

        if python_venv_name != '':
            y += 'cm run script "install python-venv" --name=' + \
                str(python_venv_name)
        else:
            y += 'cm run script "get python"'

        if python_ver != '':
            y += ' --version=' + str(python_ver)

        if python_ver_min != '':
            y += ' --version_min=' + str(python_ver_min)

    if y != '':
        x += y

    md += '```bash\n{}\n```\n'.format(x)

    st.markdown('---')
    st.markdown(md)
    st.markdown(
        '*Check [more CM installation notes at GitHub](https://github.com/mlcommons/ck/blob/master/docs/installation.md)*.')

    return {'return': 0, 'end_html': end_html}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

# Developer(s): Grigori Fursin

from cmind import utils

import os
import json
import shutil
import subprocess


def preprocess(i):

    os_info = i['os_info']
    env = i['env']
    state = i['state']
    script_path = i['run_script_input']['path']

    cm = i['automation'].cmind

    script_tags = env.get('CM_GUI_SCRIPT_TAGS', '')

    if script_tags != '':
        # Check type of tags
        if ' ' in script_tags:
            script_tags = script_tags.replace(' ', ',')

        print('Searching CM scripts using tags "{}"'.format(script_tags))

        r = cm.access({'action': 'find',
                       'automation': 'script',
                       'tags': script_tags})
        if r['return'] > 0:
            return r

        lst = r['list']

        if len(lst) == 1:
            script = lst[0]
            env['CM_GUI_SCRIPT_PATH'] = script.path
            env['CM_GUI_SCRIPT_ALIAS'] = script.meta['alias']

            print('Script found in path {}'.format(script.path))

    env['CM_GUI_SCRIPT_TAGS'] = script_tags

    # Check other vars and assemble extra CMD
    extra_cmd = env.get('CM_GUI_EXTRA_CMD', '')

    port = env.get('CM_GUI_PORT', '')
    address = env.get('CM_GUI_ADDRESS', '')
    no_browser = env.get('CM_GUI_NO_BROWSER', '')

    if no_browser != '':
        extra_cmd += ' --server.headless true'

    if address != '':
        extra_cmd += ' --server.address=' + address

    if port != '':
        extra_cmd += ' --server.port=' + port

    env['CM_GUI_EXTRA_CMD'] = extra_cmd

    print('Prepared extra CMD for streamlit: {}'.format(extra_cmd))

    return {'return': 0}



"""
dataset related classes and methods
"""

# pylint: disable=unused-argument,missing-docstring

import logging
import sys
import time

import cv2
import numpy as np


logging.basicConfig(level=logging.INFO)
log = logging.getLogger("dataset")


class Item():
    def __init__(self, label, img, idx):
        self.label = label
        self.img = img
        self.idx = idx
        self.start = time.time()


def usleep(sec):
    if sys.platform == 'win32':
        # on windows time.sleep() doesn't work to well
        import ctypes
        kernel32 = ctypes.windll.kernel32
        timer = kernel32.CreateWaitableTimerA(
            ctypes.c_void_p(), True, ctypes.c_void_p())
        delay = ctypes.c_longlong(int(-1 * (10 * 1000000 * sec)))
        kernel32.SetWaitableTimer(
            timer,
            ctypes.byref(delay),
            0,
            ctypes.c_void_p(),
            ctypes.c_void_p(),
            False)
        kernel32.WaitForSingleObject(timer, 0xffffffff)
    else:
        time.sleep(sec)


class Dataset():
    def __init__(self):
        self.arrival = None
        self.image_list = []
        self.label_list = []
        self.image_list_inmemory = {}
        self.last_loaded = -1

    def preprocess(self, use_cache=True):
        raise NotImplementedError("Dataset:preprocess")

    def get_item_count(self):
        return len(self.image_list)

    def get_list(self):
        raise NotImplementedError("Dataset:get_list")

    def load_query_samples(self, sample_list):
        self.image_list_inmemory = {}
        for sample in sample_list:
            self.image_list_inmemory[sample], _ = self.get_item(sample)
        self.last_loaded = time.time()

    def unload_query_samples(self, sample_list):
        if sample_list:
            for sample in sample_list:
                if sample in self.image_list_inmemory:
                    del self.image_list_inmemory[sample]
        else:
            self.image_list_inmemory = {}

    def get_samples(self, id_list):
        data = np.array([self.image_list_inmemory[id] for id in id_list])
        return data, self.label_list[id_list]

    def get_item_loc(self, id):
        raise NotImplementedError("Dataset:get_item_loc")


#
# Post processing
#
class PostProcessCommon:
    def __init__(self, offset=0):
        self.offset = offset
        self.good = 0
        self.total = 0

    def __call__(self, results, ids, expected=None, result_dict=None):
        processed_results = []
        n = len(results[0])
        for idx in range(0, n):
            result = results[0][idx] + self.offset
            processed_results.append([result])
            if result == expected[idx]:
                self.good += 1
        self.total += n
        return processed_results

    def add_results(self, results):
        pass

    def start(self):
        self.good = 0
        self.total = 0

    def finalize(self, results, ds=False, output_dir=None):
        results["good"] = self.good
        results["total"] = self.total


class PostProcessArgMax:
    def __init__(self, offset=0):
        self.offset = offset
        self.good = 0
        self.total = 0

    def __call__(self, results, ids, expected=None, result_dict=None):
        processed_results = []
        results = np.argmax(results[0], axis=1)
        n = results.shape[0]
        for idx in range(0, n):
            result = results[idx] + self.offset
            processed_results.append([result])
            if result == expected[idx]:
                self.good += 1
        self.total += n
        return processed_results

    def add_results(self, results):
        pass

    def start(self):
        self.good = 0
        self.total = 0

    def finalize(self, results, ds=False, output_dir=None):
        results["good"] = self.good
        results["total"] = self.total


#
# pre-processing
#

def center_crop(img, out_height, out_width):
    height, width, _ = img.shape
    left = int((width - out_width) / 2)
    right = int((width + out_width) / 2)
    top = int((height - out_height) / 2)
    bottom = int((height + out_height) / 2)
    img = img[top:bottom, left:right]
    return img


def resize_with_aspectratio(
        img, out_height, out_width, scale=87.5, inter_pol=cv2.INTER_LINEAR):
    height, width, _ = img.shape
    new_height = int(100. * out_height / scale)
    new_width = int(100. * out_width / scale)
    if height > width:
        w = new_width
        h = int(new_height * height / width)
    else:
        h = new_height
        w = int(new_width * width / height)
    img = cv2.resize(img, (w, h), interpolation=inter_pol)
    return img


def pre_process_vgg(img, dims=None, need_transpose=False):
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    output_height, output_width, _ = dims
    cv2_interpol = cv2.INTER_AREA
    img = resize_with_aspectratio(
        img,
        output_height,
        output_width,
        inter_pol=cv2_interpol)
    img = center_crop(img, output_height, output_width)
    img = np.asarray(img, dtype='float32')

    # normalize image
    means = np.array([123.68, 116.78, 103.94], dtype=np.float32)
    img -= means

    # transpose if needed
    if need_transpose:
        img = img.transpose([2, 0, 1])
    return img


def pre_process_mobilenet(img, dims=None, need_transpose=False):
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    output_height, output_width, _ = dims
    img = resize_with_aspectratio(
        img,
        output_height,
        output_width,
        inter_pol=cv2.INTER_LINEAR)
    img = center_crop(img, output_height, output_width)
    img = np.asarray(img, dtype='float32')

    img /= 255.0
    img -= 0.5
    img *= 2

    # transpose if needed
    if need_transpose:
        img = img.transpose([2, 0, 1])
    return img


def pre_process_imagenet_pytorch(img, dims=None, need_transpose=False):
    from PIL import Image
    import torchvision.transforms.functional as F

    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img = Image.fromarray(img)
    img = F.resize(img, 256, Image.BILINEAR)
    img = F.center_crop(img, 224)
    img = F.to_tensor(img)
    img = F.normalize(
        img, mean=[
            0.485, 0.456, 0.406], std=[
            0.229, 0.224, 0.225], inplace=False)
    if not need_transpose:
        img = img.permute(1, 2, 0)  # NHWC
    img = np.asarray(img, dtype='float32')
    return img


def maybe_resize(img, dims):
    img = np.array(img, dtype=np.float32)
    if len(img.shape) < 3 or img.shape[2] != 3:
        # some images might be grayscale
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    if dims is not None:
        im_height, im_width, _ = dims
        img = cv2.resize(img, (im_width, im_height),
                         interpolation=cv2.INTER_LINEAR)
    return img


def pre_process_coco_mobilenet(img, dims=None, need_transpose=False):
    img = maybe_resize(img, dims)
    img = np.asarray(img, dtype=np.uint8)
    # transpose if needed
    if need_transpose:
        img = img.transpose([2, 0, 1])
    return img


def pre_process_coco_pt_mobilenet(img, dims=None, need_transpose=False):
    img = maybe_resize(img, dims)
    img -= 127.5
    img /= 127.5
    # transpose if needed
    if need_transpose:
        img = img.transpose([2, 0, 1])
    return img


def pre_process_coco_resnet34(img, dims=None, need_transpose=False):
    img = maybe_resize(img, dims)
    mean = np.array([0.485, 0.456, 0.406], dtype=np.float32)
    std = np.array([0.229, 0.224, 0.225], dtype=np.float32)

    img = img / 255. - mean
    img = img / std

    if need_transpose:
        img = img.transpose([2, 0, 1])

    return img


def pre_process_coco_resnet34_tf(img, dims=None, need_transpose=False):
    img = maybe_resize(img, dims)
    mean = np.array([123.68, 116.78, 103.94], dtype=np.float32)
    img = img - mean
    if need_transpose:
        img = img.transpose([2, 0, 1])

    return img


def pre_process_openimages_retinanet(img, dims=None, need_transpose=False):
    img = maybe_resize(img, dims)
    img /= 255.
    # transpose if needed
    if need_transpose:
        img = img.transpose([2, 0, 1])
    return img



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os
import shutil


def preprocess(i):

    env = i['env']

    if str(env.get('CM_DATASET_PREPROCESSED_BY_MLC', '')
           ).lower() in ["yes", "1", "true"]:
        run_dir = os.getcwd()
        if env.get('CM_DATASET_CALIBRATION', '') == "yes":
            env['CM_DATASET_CALIBRATION_PATH'] = os.path.join(
                env['CM_OPENORCA_PREPROCESSED_ROOT'],
                "open_orca_gpt4_tokenized_llama.calibration_1000.pkl.gz")
            env['CM_GET_DEPENDENT_CACHED_PATH'] = env['CM_DATASET_CALIBRATION_PATH']
            env['CM_DATASET_OPENORCA_CALIBRATION_PATH'] = env['CM_DATASET_CALIBRATION_PATH']
        else:
            env['CM_DATASET_PREPROCESSED_PATH'] = os.path.join(
                env['CM_OPENORCA_PREPROCESSED_ROOT'],
                "open_orca_gpt4_tokenized_llama.sampled_24576.pkl.gz")
            env['CM_GET_DEPENDENT_CACHED_PATH'] = env['CM_DATASET_PREPROCESSED_PATH']
            env['CM_DATASET_OPENORCA_PREPROCESSED_PATH'] = env['CM_DATASET_PREPROCESSED_PATH']
        # run_cmd = f"gunzip -k {env['CM_DATASET_PREPROCESSED_PATH']}"
        run_cmd = ''
    else:
        inference_src = env['CM_MLPERF_INFERENCE_SOURCE']
        run_dir = os.path.join(inference_src, 'language', 'llama2-70b')
        model_dir = env['CM_ML_MODEL_PATH']
        if env.get('CM_DATASET_CALIBRATION', '') == "yes":
            return {'return': 1, 'error': 'No raw preprocessing information is available for openorca calibration. Please use _mlcommons variation to use the MLCommons shared calibration dataset'}
        else:
            env['CM_DATASET_PREPROCESSED_PATH'] = os.path.join(
                os.path.join(
                    os.getcwd(),
                    "processed-openorca",
                    'open_orca_gpt4_tokenized_llama.sampled_' +
                    env['CM_DATASET_SIZE'] +
                    '.pkl'))
            run_cmd = env['CM_PYTHON_BIN_WITH_PATH'] + ' processorca.py --dataset_pq_path=' + env['CM_DATASET_OPENORCA_PARQUET'] + ' --model_dir=' + model_dir + \
                ' --seqlen_limit=2048 --export_dir=' + \
                os.path.join(os.getcwd(), "processed-openorca") + \
                ' --num_total_samples=' + env['CM_DATASET_SIZE']

    env['CM_RUN_DIR'] = run_dir
    env['CM_RUN_CMD'] = run_cmd

    return {'return': 0}


def postprocess(i):
    env = i['env']

    return {'return': 0}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os
import json
import shutil
import cmind
import sys
from tabulate import tabulate
import mlperf_utils


def preprocess(i):
    return {'return': 0}

# Helper function to fill dictionary from JSON file


def fill_from_json(file_path, keys, sut_info):
    with open(file_path, 'r') as f:
        data = json.load(f)
        for key in keys:
            if key in data and (
                    sut_info[key] is None or sut_info[key] == "default"):
                sut_info[key] = data[key]
            elif key in data and sut_info[key] != data[key]:
                return -1  # error saying there is a mismatch in the value of a key
        return sut_info

# Helper function to check whether all the keys(sut information) are assigned


def check_dict_filled(keys, sut_info):
    for key in keys:
        if key in sut_info and sut_info[key] is None:
            return False
    return True

# The function checks whether the submitting model name belongs standard
# model names for MLPef Inference


def model_in_valid_models(model, mlperf_version):
    import submission_checker as checker
    config = checker.MODEL_CONFIG

    if model not in config[mlperf_version]['models']:
        internal_model_name = config[mlperf_version]["model_mapping"].get(
            model, '')  # resnet50 -> resnet
        if internal_model_name == '':
            # Indicate failure with no internal model name
            return (False, None)
        else:
            # Indicate success with internal model name
            return (True, internal_model_name)
    else:
        return (True, model)


def generate_submission(env, state, inp, submission_division):

    # Save current user directory
    cur_dir = os.getcwd()

    if env.get('CM_MLPERF_INFERENCE_RESULTS_DIR_', '') == '':
        results_dir = os.path.join(
            env['CM_MLPERF_INFERENCE_RESULTS_DIR'],
            f"{env['CM_MLPERF_RUN_STYLE']}_results")
    else:
        results_dir = env['CM_MLPERF_INFERENCE_RESULTS_DIR_']

    mlperf_path = env['CM_MLPERF_INFERENCE_SOURCE']
    submission_checker_dir = os.path.join(mlperf_path, "tools", "submission")
    sys.path.append(submission_checker_dir)

    if env.get('CM_MLPERF_INFERENCE_SUBMISSION_DIR', '') == '':
        from pathlib import Path
        user_home = str(Path.home())
        env['CM_MLPERF_INFERENCE_SUBMISSION_DIR'] = os.path.join(
            user_home, "mlperf_submission")

    submission_dir = env.get('CM_MLPERF_INFERENCE_SUBMISSION_DIR', '')
    if submission_dir == '':
        submission_base_dir = env.get(
            'CM_MLPERF_INFERENCE_SUBMISSION_BASE_DIR', '')
        if submission_base_dir == '':
            return {'return': 1, 'error': f"Both CM_MLPERF_INFERENCE_SUBMISSION_DIR and CM_MLPERF_INFERENCE_SUBMISSION_BASE_DIR can not be empty!"}
        else:
            submission_dir = os.path.join(
                submission_base_dir, "mlperf_inference_submission")
            env['CM_MLPERF_INFERENCE_SUBMISSION_DIR'] = submission_dir

    if env.get('CM_MLPERF_CLEAN_SUBMISSION_DIR', '') != '':
        print('=================================================')
        print(
            'Cleaning {} ...'.format(
                env['CM_MLPERF_INFERENCE_SUBMISSION_DIR']))
        if os.path.exists(submission_dir):
            shutil.rmtree(submission_dir)
        print('=================================================')

    if not os.path.isdir(submission_dir):
        os.makedirs(submission_dir)

    if str(env.get('CM_MLPERF_SUBMISSION_DIR_SHARED', '')
           ).lower() in ["yes", "true", "1"]:
        os.chmod(submission_dir, 0o2775)

    print('* MLPerf inference submission dir: {}'.format(submission_dir))
    print('* MLPerf inference results dir: {}'.format(results_dir))
    results = [
        f for f in os.listdir(results_dir) if not os.path.isfile(
            os.path.join(
                results_dir,
                f))]

    system_meta_default = state['CM_SUT_META']

    # set pytorch as the default framework
    if system_meta_default['framework'] == '':
        system_meta_default['framework'] = "pytorch"

    system_meta_tmp = {}
    if 'CM_MLPERF_SUBMISSION_SYSTEM_TYPE' in env:
        system_meta_tmp['system_type'] = env['CM_MLPERF_SUBMISSION_SYSTEM_TYPE']

    if submission_division != "":
        system_meta_tmp['division'] = submission_division
        division = submission_division
    else:
        division = system_meta_default['division']

    if 'CM_MLPERF_SUBMISSION_CATEGORY' in env:
        system_meta_tmp['system_type'] = env['CM_MLPERF_SUBMISSION_CATEGORY'].replace(
            "-", ",")

    duplicate = (
        env.get(
            'CM_MLPERF_DUPLICATE_SCENARIO_RESULTS',
            'no') in [
            "yes",
            "True"])

    if division not in ['open', 'closed']:
        return {'return': 1, 'error': '"division" must be "open" or "closed"'}

    print('* MLPerf inference division: {}'.format(division))

    path_submission_root = submission_dir
    path_submission_division = os.path.join(path_submission_root, division)
    if not os.path.isdir(path_submission_division):
        os.makedirs(path_submission_division)

    # Check submitter
    if env.get('CM_MLPERF_SUBMITTER'):
        submitter = env['CM_MLPERF_SUBMITTER']
        system_meta_tmp['submitter'] = submitter
    else:
        submitter = system_meta_default['submitter']
        env['CM_MLPERF_SUBMITTER'] = submitter

    print('* MLPerf inference submitter: {}'.format(submitter))

    if env.get('CM_MLPERF_SUT_SW_NOTES_EXTRA', '') != '':
        sw_notes = f"""{
            system_meta_tmp['sw_notes']} {
            env['CM_MLPERF_SUT_SW_NOTES_EXTRA']}"""
        system_meta_tmp['sw_notes'] = sw_notes

    if env.get('CM_MLPERF_SUT_HW_NOTES_EXTRA', '') != '':
        hw_notes = f"""{
            system_meta_tmp['hw_notes']} {
            env['CM_MLPERF_SUT_HW_NOTES_EXTRA']}"""
        system_meta_tmp['hw_notes'] = hw_notes

    path_submission = os.path.join(path_submission_division, submitter)
    if not os.path.isdir(path_submission):
        os.makedirs(path_submission)

    # SUT base
    system = env.get('CM_HW_NAME', 'default').replace(' ', '_')

    code_path = os.path.join(path_submission, "code")

    for res in results:
        system_meta = {}
        system_meta.update(system_meta_tmp)
        result_path = os.path.join(results_dir, res)
        # variable to check whether the sut_meta.json is present in the root
        # folder
        saved_system_meta_file_path = os.path.join(
            result_path, 'system_meta.json')
        # checks for json file containing system meta
        sut_info = {
            "system_name": None,
            "implementation": None,
            "device": None,
            "framework": None,
            "framework_version": "default",
            "run_config": "default"
        }  # variable to store the system meta

        model_mapping_combined = {}  # to store all the model mapping related to an SUT

        # check whether the root folder contains the sut infos
        # if yes then there is no need to check for meta files inside
        # individual model folders
        if "cm-sut-info.json" in os.listdir(result_path):
            sut_info = fill_from_json(
                os.path.join(
                    result_path,
                    "cm-sut-info.json"),
                sut_info.keys(),
                sut_info)
            if sut_info == -1:
                return {
                    'return': 1, 'error': f"key value mismatch. Refer the populating dictionary:\n{sut_info}\n and file {os.path.join(result_path, 'cm-sut-info.json')}"}
            if check_dict_filled(sut_info.keys(), sut_info):
                print(
                    f"sut info completely filled from {os.path.join(result_path, 'cm-sut-info.json')}!")

        # Check whether the root folder contains the model mapping file
        # expects json file in the format:
        # {
        #   custom_model1:official_model(could be any official model),
        #   custom_model2:official_model(could be any official model)
        # }
        if "model_mapping.json" in os.listdir(result_path):
            with open(os.path.join(result_path, "model_mapping.json"), 'r') as f:
                model_mapping_combined = json.load(f)

        # Preprocessing part.
        # Even the model mapping json file is present in root directory, the folders are traversed
        # and the data is updated provided not duplicated.
        models = [
            f for f in os.listdir(result_path) if not os.path.isfile(
                os.path.join(
                    result_path, f))]
        if division == "open" and len(model_mapping_combined) == 0:
            for model in models:
                is_valid, returned_model_name = model_in_valid_models(
                    model, env.get('CM_MLPERF_LAST_RELEASE', 'v4.1'))
                if not is_valid:
                    result_model_path = os.path.join(result_path, model)
                    scenarios = [
                        f for f in os.listdir(result_model_path) if not os.path.isfile(
                            os.path.join(
                                result_model_path, f))]
                    for scenario in scenarios:
                        result_scenario_path = os.path.join(
                            result_model_path, scenario)
                        modes = [
                            f for f in os.listdir(result_scenario_path) if not os.path.isfile(
                                os.path.join(
                                    result_scenario_path, f))]
                        for mode in modes:
                            result_mode_path = os.path.join(
                                result_scenario_path, mode)
                            if mode == "performance":
                                compliance_performance_run_path = os.path.join(
                                    result_mode_path, "run_1")
                                # model mapping part
                                tmp_model_mapping_file_path = os.path.join(
                                    compliance_performance_run_path, "model_mapping.json")
                                if os.path.exists(tmp_model_mapping_file_path):
                                    with open(tmp_model_mapping_file_path, 'r') as f:
                                        new_model_mapping = json.load(f)
                                        for new_custom_model in new_model_mapping:
                                            if new_custom_model not in model_mapping_combined:
                                                model_mapping_combined.update(
                                                    {new_custom_model: new_model_mapping[new_custom_model]})
                                else:
                                    return {
                                        "return": 1, "error": f"model_mapping.json not found in {compliance_performance_run_path}"}
                else:
                    if returned_model_name != model:
                        model_mapping_combined.update(
                            {model: returned_model_name})

        if check_dict_filled(sut_info.keys(), sut_info):
            system = env.get('CM_HW_NAME', sut_info["system_name"])
            implementation = sut_info["implementation"]
            device = sut_info["device"]
            framework = sut_info["framework"].replace(" ", "_")
            framework_version = sut_info["framework_version"]
            run_config = sut_info["run_config"]
            new_res = f"{system}-{implementation}-{device}-{framework}-{run_config}"
        else:
            new_res = res

        print(f"The SUT folder name for submission generation is: {new_res}")

        platform_prefix = inp.get('platform_prefix', '')
        if platform_prefix:
            sub_res = platform_prefix + "-" + new_res
        else:
            sub_res = new_res

        submission_path = os.path.join(path_submission, "results", sub_res)
        measurement_path = os.path.join(
            path_submission, "measurements", sub_res)
        compliance_path = os.path.join(path_submission, "compliance", sub_res)
        system_path = os.path.join(path_submission, "systems")
        submission_system_path = system_path

        if not os.path.isdir(submission_path):
            os.makedirs(submission_path)
        if not os.path.isdir(measurement_path):
            os.makedirs(measurement_path)
        if not os.path.isdir(submission_system_path):
            os.makedirs(submission_system_path)
        system_file = os.path.join(submission_system_path, sub_res + ".json")

        # Save the model mapping json file
        if model_mapping_combined:
            with open(os.path.join(path_submission, "model_mapping.json"), "w") as fp:
                json.dump(model_mapping_combined, fp, indent=2)

        models = [
            f for f in os.listdir(result_path) if not os.path.isfile(
                os.path.join(
                    result_path, f))]

        results = {}

        model_platform_info_file = None

        for model in models:
            results[model] = {}
            platform_info_file = None
            result_model_path = os.path.join(result_path, model)
            submission_model_path = os.path.join(submission_path, model)
            measurement_model_path = os.path.join(measurement_path, model)
            compliance_model_path = os.path.join(compliance_path, model)
            code_model_path = os.path.join(code_path, model)
            scenarios = [
                f for f in os.listdir(result_model_path) if not os.path.isfile(
                    os.path.join(
                        result_model_path, f))]
            submission_code_path = code_model_path
            if not os.path.isdir(submission_code_path):
                os.makedirs(submission_code_path)
            if not os.path.exists(os.path.join(
                    submission_code_path, "README.md")):
                with open(os.path.join(submission_code_path, "README.md"), mode='w') as f:
                    f.write("TBD")  # create an empty README

            print('* MLPerf inference model: {}'.format(model))
            for scenario in scenarios:
                # the system_info.txt is copied from the mode directory if
                # found, else it would be looked under scenario directory
                system_info_from_mode = False
                results[model][scenario] = {}
                result_scenario_path = os.path.join(
                    result_model_path, scenario)
                submission_scenario_path = os.path.join(
                    submission_model_path, scenario)
                measurement_scenario_path = os.path.join(
                    measurement_model_path, scenario)
                compliance_scenario_path = os.path.join(
                    compliance_model_path, scenario)

                if duplicate and scenario == 'singlestream':
                    if not os.path.exists(os.path.join(
                            result_model_path, "offline")):
                        print(
                            'Duplicating results from {} to offline:'.format(scenario))
                        shutil.copytree(
                            result_scenario_path, os.path.join(
                                result_model_path, "offline"))
                        scenarios.append("offline")
                    if not os.path.exists(os.path.join(
                            result_model_path, "multistream")):
                        print(
                            'Duplicating results from {} to multistream:'.format(scenario))
                        shutil.copytree(
                            result_scenario_path, os.path.join(
                                result_model_path, "multistream"))
                        scenarios.append("multistream")

                modes = [
                    f for f in os.listdir(result_scenario_path) if not os.path.isfile(
                        os.path.join(
                            result_scenario_path, f))]
                power_run = False

                # we check for the existance of mlperf_log_summary.txt
                # mlperf_log_detail.txt to consider a result folder as valid.
                # Rest of the checks are done later by the submission checker
                files_to_check = [
                    "mlperf_log_summary.txt",
                    "mlperf_log_detail.txt"]
                if not all([os.path.exists(os.path.join(
                        result_scenario_path, "performance", "run_1", f)) for f in files_to_check]):
                    continue

                if not os.path.isdir(measurement_scenario_path):
                    os.makedirs(measurement_scenario_path)

                for mode in modes:
                    result_mode_path = os.path.join(result_scenario_path, mode)
                    submission_mode_path = os.path.join(
                        submission_scenario_path, mode)
                    submission_measurement_path = measurement_scenario_path
                    submission_compliance_path = os.path.join(
                        compliance_scenario_path, mode)
                    if mode.startswith("TEST"):
                        submission_results_path = submission_compliance_path
                    else:
                        submission_results_path = submission_mode_path
                    if os.path.exists(submission_results_path):
                        shutil.rmtree(submission_results_path)

                    if mode == 'performance':

                        if os.path.exists(os.path.join(
                                result_mode_path, "power")):
                            power_run = True
                            result_power_path = os.path.join(
                                result_mode_path, 'power')
                            submission_power_path = os.path.join(
                                submission_mode_path, 'power')
                            os.makedirs(submission_power_path)
                            power_files = []
                            for f in os.listdir(result_power_path):
                                # Todo add required check from
                                # submission_checker
                                power_files.append(f)
                            for f in power_files:
                                shutil.copy(
                                    os.path.join(
                                        result_power_path, f), os.path.join(
                                        submission_power_path, f))

                            analyzer_settings_file = env.get(
                                'CM_MLPERF_POWER_ANALYZER_SETTINGS_FILE_PATH', os.path.join(
                                    env['CM_TMP_CURRENT_SCRIPT_PATH'], "default_files", "analyzer_table.md"))
                            power_settings_file = env.get(
                                'CM_MLPERF_POWER_SETTINGS_FILE_PATH', os.path.join(
                                    env['CM_TMP_CURRENT_SCRIPT_PATH'], "default_files", "power_settings.md"))

                            shutil.copy(
                                analyzer_settings_file, os.path.join(
                                    submission_measurement_path, "analyzer_table.md"))
                            shutil.copy(
                                power_settings_file, os.path.join(
                                    submission_measurement_path, "power_settings.md"))

                            result_ranging_path = os.path.join(
                                result_mode_path, 'ranging')
                            submission_ranging_path = os.path.join(
                                submission_mode_path, 'ranging')
                            os.makedirs(submission_ranging_path)
                            ranging_files = []
                            for f in os.listdir(result_ranging_path):
                                # Todo add required check from
                                # submission_checker
                                ranging_files.append(f)
                            for f in ranging_files:
                                shutil.copy(
                                    os.path.join(
                                        result_ranging_path, f), os.path.join(
                                        submission_ranging_path, f))

                        result_mode_path = os.path.join(
                            result_mode_path, 'run_1')
                        submission_results_path = os.path.join(
                            submission_mode_path, 'run_1')

                        if not os.path.exists(saved_system_meta_file_path):
                            if os.path.exists(os.path.join(
                                    result_mode_path, "system_meta.json")):
                                saved_system_meta_file_path = os.path.join(
                                    result_mode_path, "system_meta.json")
                            else:
                                print("WARNING: system_meta.json was not found in the SUT root or mode directory inside the results folder. CM is automatically creating one using the system defaults. Please modify them as required.")
                        if os.path.exists(saved_system_meta_file_path):
                            with open(saved_system_meta_file_path, "r") as f:
                                saved_system_meta = json.load(f)
                                for key in list(saved_system_meta):
                                    if saved_system_meta[key] is None or str(
                                            saved_system_meta[key]).strip() == '':
                                        del (saved_system_meta[key])
                                if saved_system_meta["division"] != "" and submission_division == "":
                                    system_meta["division"] = saved_system_meta["division"]
                                # override the saved meta with the user inputs
                                system_meta = {
                                    **saved_system_meta, **system_meta}
                        # add any missing fields from the defaults, if
                        # system_meta.json is not detected, default one will be
                        # written
                        system_meta = {**system_meta_default, **system_meta}
                        print(system_meta)
                        # check if framework version is there in system_meta,
                        # if not try to fill it from sut_info
                        if system_meta['framework'] == "":
                            system_meta['framework'] = sut_info.get(
                                'framework', '') + sut_info.get('framework_version', '')
                            if system_meta['framework'] == "":
                                print(
                                    "WARNING: framework field could not be filled from system_meta.json or sut_info.json. This will trigger error in submission checker")

                    if not os.path.isdir(submission_results_path):
                        os.makedirs(submission_results_path)

                    # if division == "closed" and not os.path.isdir(submission_compliance_path):
                    #    os.makedirs(submission_compliance_path)

                    user_conf_path = os.path.join(
                        result_scenario_path, "user.conf")
                    if os.path.exists(user_conf_path):
                        shutil.copy(
                            user_conf_path, os.path.join(
                                measurement_scenario_path, 'user.conf'))
                    else:
                        user_conf_path = os.path.join(
                            result_mode_path, "user.conf")
                        if os.path.exists(user_conf_path):
                            shutil.copy(
                                user_conf_path, os.path.join(
                                    submission_measurement_path, 'user.conf'))
                        else:
                            if mode.lower() == "performance":
                                return {
                                    "return": 1, "error": f"user.conf missing in both paths: {user_conf_path} and {os.path.join(result_scenario_path, 'user.conf')}"}

                    measurements_json_path = os.path.join(
                        result_scenario_path, "measurements.json")
                    target_measurement_json_path = measurement_scenario_path
                    if not os.path.exists(measurements_json_path):
                        measurements_json_path = os.path.join(
                            result_mode_path, "measurements.json")
                        target_measurement_json_path = submission_measurement_path

                    if os.path.exists(measurements_json_path):
                        with open(measurements_json_path, "r") as f:
                            measurements_json = json.load(f)
                            model_precision = measurements_json.get(
                                "weight_data_types", "fp32")
                        shutil.copy(
                            measurements_json_path,
                            os.path.join(
                                target_measurement_json_path,
                                sub_res + '.json'))
                        shutil.copy(
                            measurements_json_path,
                            os.path.join(
                                target_measurement_json_path,
                                'model-info.json'))
                    else:
                        if mode.lower() == "performance":
                            return {
                                "return": 1, "error": f"measurements.json missing in both paths: {measurements_json_path} and {os.path.join(result_scenario_path, 'user.conf')}"}

                    files = []
                    readme = False

                    for f in os.listdir(result_mode_path):
                        if mode.startswith("TEST"):
                            if f.startswith('verify_'):
                                files.append(f)
                            elif f == "performance":
                                compliance_performance_run_path = os.path.join(
                                    result_mode_path, f, "run_1")
                                if os.path.exists(
                                        compliance_performance_run_path):
                                    target = os.path.join(
                                        submission_results_path, "performance", "run_1")
                                    os.makedirs(target)
                                    for log_file in os.listdir(
                                            compliance_performance_run_path):
                                        if log_file.startswith("mlperf_"):
                                            shutil.copy(
                                                os.path.join(
                                                    compliance_performance_run_path, log_file), os.path.join(
                                                    target, log_file))
                            elif f == "accuracy":
                                compliance_accuracy_run_path = os.path.join(
                                    result_mode_path, f)
                                if os.path.exists(
                                        compliance_accuracy_run_path):
                                    target = os.path.join(
                                        submission_results_path, "accuracy")
                                    os.makedirs(target)
                                    for log_file in os.listdir(
                                            compliance_accuracy_run_path):
                                        log_file_name = os.path.basename(
                                            log_file)
                                        # print(os.path.join(compliance_accuracy_run_path, log_file))
                                        if log_file_name in [
                                                "mlperf_log_accuracy.json", "accuracy.txt", "baseline_accuracy.txt", "compliance_accuracy.txt"]:
                                            shutil.copy(
                                                os.path.join(
                                                    compliance_accuracy_run_path, log_file), os.path.join(
                                                    target, log_file))
                        else:
                            if f.startswith('mlperf_') and not f.endswith(
                                    'trace.json'):
                                files.append(f)
                            elif f == "spl.txt":
                                files.append(f)
                            elif f in ["README.md", "README-extra.md", "cm-version-info.json", "os_info.json", "cpu_info.json", "pip_freeze.json", "system_info.txt", "cm-deps.png", "cm-deps.mmd"] and mode == "performance":
                                shutil.copy(
                                    os.path.join(
                                        result_mode_path, f), os.path.join(
                                        submission_measurement_path, f))
                                if f == "system_info.txt" and not platform_info_file:
                                    # the first found system_info.txt will be taken as platform info file for a specific model to be placed in
                                    # measurements-model folder when generating
                                    # the final submission
                                    platform_info_file = os.path.join(
                                        result_mode_path, f)
                            elif f in ["console.out"]:
                                shutil.copy(
                                    os.path.join(
                                        result_mode_path, f), os.path.join(
                                        submission_measurement_path, mode + "_" + f))

                    if mode == "accuracy":
                        if os.path.exists(os.path.join(
                                result_mode_path, "accuracy.txt")):
                            files.append("accuracy.txt")
                        if model == "stable-diffusion-xl" and os.path.exists(
                                os.path.join(result_mode_path, "images")):
                            shutil.copytree(
                                os.path.join(
                                    result_mode_path, "images"), os.path.join(
                                    submission_results_path, "images"))

                    for f in files:
                        print(' * ' + f)
                        p_target = os.path.join(submission_results_path, f)
                        shutil.copy(
                            os.path.join(
                                result_mode_path,
                                f),
                            p_target)

                if os.path.exists(os.path.join(
                        result_scenario_path, "system_info.txt")):
                    shutil.copy(
                        os.path.join(
                            result_scenario_path, "system_info.txt"), os.path.join(
                            submission_measurement_path, f))
                    platform_info_file = os.path.join(
                        result_scenario_path, "system_info.txt")

                readme_file = os.path.join(
                    submission_measurement_path, "README.md")
                if not os.path.exists(readme_file):
                    with open(readme_file, mode='w') as f:
                        f.write("TBD")  # create an empty README

                readme_suffix = ""
                result_string, result = mlperf_utils.get_result_string(
                    env['CM_MLPERF_LAST_RELEASE'], model, scenario, result_scenario_path, power_run, sub_res, division, system_file, model_precision, env.get('CM_MLPERF_INFERENCE_SOURCE_VERSION'))

                for key in result:
                    results[model][scenario][key] = result[key]
                with open(readme_file, mode='a') as f:
                    f.write(result_string)

            # Copy system_info.txt to the submission measurements model folder
            # if any scenario performance run has it
            sys_info_file = None

            if os.path.exists(os.path.join(
                    result_model_path, "system_info.txt")):
                sys_info_file = os.path.join(
                    result_model_path, "system_info.txt")
            elif platform_info_file:
                sys_info_file = platform_info_file

            if sys_info_file:
                model_platform_info_file = sys_info_file
                shutil.copy(
                    sys_info_file,
                    os.path.join(
                        measurement_model_path,
                        "system_info.txt"))

        # Copy system_info.txt to the submission measurements folder if any
        # model performance run has it
        sys_info_file = None

        if os.path.exists(os.path.join(result_path, "system_info.txt")):
            sys_info_file = os.path.join(result_path, "system_info.txt")
        elif model_platform_info_file:
            sys_info_file = model_platform_info_file

        if sys_info_file:
            shutil.copy(
                sys_info_file,
                os.path.join(
                    measurement_path,
                    "system_info.txt"))
        else:
            if env.get('CM_GET_PLATFORM_DETAILS', '') == "yes":
                cm_input = {'action': 'run',
                            'automation': 'script',
                            'tags': 'get,platform,details',
                            'env': {'CM_PLATFORM_DETAILS_FILE_PATH': os.path.join(measurement_path, "system_info.txt")},
                            'quiet': True
                            }
                r = cmind.access(cm_input)
                if r['return'] > 0:
                    return r

        with open(system_file, "w") as fp:
            json.dump(system_meta, fp, indent=2)

        result_table, headers = mlperf_utils.get_result_table(results)

        print(tabulate(result_table, headers=headers, tablefmt="pretty"))

        sut_readme_file = os.path.join(measurement_path, "README.md")
        with open(sut_readme_file, mode='w') as f:
            f.write(tabulate(result_table, headers=headers, tablefmt="github"))

    return {'return': 0}


def postprocess(i):
    env = i['env']
    state = i['state']
    inp = i['input']

    submission_divisions = []

    if env.get('CM_MLPERF_SUBMISSION_DIVISION', '') in [
            "open-closed", "closed-open"]:
        submission_divisions = ["open", "closed"]
    elif env.get('CM_MLPERF_SUBMISSION_DIVISION', '') != '':
        submission_divisions.append(env['CM_MLPERF_SUBMISSION_DIVISION'])

    # if submission division is not assigned, default value would be taken in
    # submission_generation function
    if env.get('CM_MLPERF_SUBMISSION_DIVISION', '') == '':
        r = generate_submission(env, state, inp, submission_division="")
        if r['return'] > 0:
            return r
    else:
        for submission_division in submission_divisions:
            r = generate_submission(env, state, inp, submission_division)
            if r['return'] > 0:
                return r

    return {'return': 0}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os


def preprocess(i):

    os_info = i['os_info']

    env = i['env']

    recursion_spaces = i['recursion_spaces']

    file_name = 'go.exe' if os_info['platform'] == 'windows' else 'go'
    env['FILE_NAME'] = file_name
    if 'CM_GO_BIN_WITH_PATH' not in env:
        r = i['automation'].find_artifact({'file_name': file_name,
                                           'env': env,
                                           'os_info': os_info,
                                           'default_path_env_key': 'PATH',
                                           'detect_version': True,
                                           'env_path_key': 'CM_GO_BIN_WITH_PATH',
                                           'run_script_input': i['run_script_input'],
                                           'recursion_spaces': recursion_spaces})
        if r['return'] > 0:
            if r['return'] == 16:
                env['CM_REQUIRE_INSTALL'] = "yes"
                return {'return': 0}
            else:
                return r

    return {'return': 0}


def detect_version(i):
    r = i['automation'].parse_version({'match_text': r'\s+go([\d.]+)',
                                       'group_number': 1,
                                       'env_key': 'CM_GO_VERSION',
                                       'which_env': i['env']})
    if r['return'] > 0:
        return r

    version = r['version']

    print(i['recursion_spaces'] + '    Detected version: {}'.format(version))
    return {'return': 0, 'version': version}


def postprocess(i):
    env = i['env']

    r = detect_version(i)

    if r['return'] > 0:
        return r

    version = r['version']
    found_file_path = env['CM_GO_BIN_WITH_PATH']

    found_path = os.path.dirname(found_file_path)
    env['CM_GO_INSTALLED_PATH'] = found_path

    env['CM_GO_CACHE_TAGS'] = 'version-' + version

    return {'return': 0, 'version': version}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os
import configparser


def preprocess(i):

    os_info = i['os_info']

    env = i['env']

    recursion_spaces = i['recursion_spaces']

    file_name = 'rclone.exe' if os_info['platform'] == 'windows' else 'rclone'
    env['FILE_NAME'] = file_name

    run_script_input = i['run_script_input']
    automation = i['automation']

    need_version = env.get('CM_VERSION', '')

    host_os_machine = ''
    if os_info['platform'] != 'windows':
        host_os_machine = env['CM_HOST_OS_MACHINE']  # ABI

    r = automation.detect_version_using_script({
        'env': env,
        'run_script_input': run_script_input,
        'recursion_spaces': recursion_spaces})

    if r['return'] > 0:
        if r['return'] == 16:
            install_script = 'install'
            if os_info['platform'] != 'windows' and env.get(
                    'CM_RCLONE_SYSTEM', '') == 'yes':
                install_script += '-system'
            else:
                if os_info['platform'] != 'windows':
                    x1 = 'arm64' if host_os_machine.startswith(
                        'arm') or host_os_machine.startswith('aarch') else 'amd64'

                    filebase = 'rclone-v{}-{}-{}'
                    urlbase = 'https://downloads.rclone.org/v{}/{}'

                    if os_info['platform'] == 'darwin':
                        filename = filebase.format(need_version, 'osx', x1)
                    elif os_info['platform'] == 'linux':
                        filename = filebase.format(need_version, 'linux', x1)

                    env['CM_RCLONE_URL'] = urlbase.format(
                        need_version, filename + '.zip')
                    env['CM_RCLONE_ARCHIVE'] = filename
                    env['CM_RCLONE_ARCHIVE_WITH_EXT'] = filename + '.zip'

                    print(
                        recursion_spaces +
                        'Downloading {}'.format(
                            env['CM_RCLONE_URL']))

                cur_dir = os.getcwd()
                path_bin = os.path.join(cur_dir, file_name)
                env['CM_RCLONE_BIN_WITH_PATH'] = path_bin

                if not env.get('+PATH', []):
                    env['+PATH'] = []
                env['+PATH'].append(cur_dir)

                if not env.get('+PATH', []):
                    env['+PATH'] = []
                env['+PATH'].append(cur_dir)

            r = automation.run_native_script({'run_script_input': run_script_input,
                                              'env': env,
                                              'script_name': install_script})
            if r['return'] > 0:
                return r
        else:
            return r

    return {'return': 0}


def detect_version(i):
    r = i['automation'].parse_version({'match_text': r'rclone v([\d.]+)',
                                       'group_number': 1,
                                       'env_key': 'CM_RCLONE_VERSION',
                                       'which_env': i['env']})
    if r['return'] > 0:
        return r

    version = r['version']

    print(i['recursion_spaces'] + '    Detected version: {}'.format(version))

    return {'return': 0, 'version': version}


def postprocess(i):

    os_info = i['os_info']
    env = i['env']

    gdrive = env.get('CM_RCLONE_GDRIVE', '')
    if gdrive == "yes":
        config = configparser.ConfigParser()
        config_file_path = os.path.join(
            env['CM_TMP_CURRENT_SCRIPT_PATH'], "configs", "rclone.conf")

        config.read(config_file_path)
        # config['cm-team']['service_account_file'] = os.path.join(env['CM_TMP_CURRENT_SCRIPT_PATH'], "accessfiles", "rclone-gdrive.json")

        default_config_path = os.path.join(
            os.path.expanduser('~'), ".config", "rclone", "rclone.conf")

        default_config = configparser.ConfigParser()
        default_config.read(default_config_path)

        for section in config.sections():
            if section not in default_config.sections():
                default_config[section] = config[section]

        with open(default_config_path, 'w') as configfile:
            default_config.write(configfile)
        print({section: dict(default_config[section])
              for section in default_config.sections()})

    r = detect_version(i)

    if r['return'] > 0:
        return r

    version = r['version']

    env['CM_RCLONE_CACHE_TAGS'] = 'version-' + version

    file_name = 'rclone.exe' if os_info['platform'] == 'windows' else 'rclone'

    if os_info['platform'] == 'windows' or env.get(
            'CM_RCLONE_SYSTEM', '') != 'yes':
        cur_dir = os.getcwd()
        path_bin = os.path.join(cur_dir, file_name)
        if os.path.isfile(path_bin):
            # Was downloaded and extracted by CM
            env['CM_RCLONE_BIN_WITH_PATH'] = path_bin
            env['+PATH'] = [cur_dir]

    return {'return': 0, 'version': version}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import os


def preprocess(i):

    os_info = i['os_info']

    env = i['env']

    recursion_spaces = i['recursion_spaces']

    file_name = 'gh.exe' if os_info['platform'] == 'windows' else 'gh'

    # Will check env['CM_TMP_PATH'] if comes from installation script
    r = i['automation'].find_artifact({'file_name': file_name,
                                       'env': env,
                                       'os_info': os_info,
                                       'default_path_env_key': 'PATH',
                                       'detect_version': True,
                                       'env_path_key': 'CM_GITHUBCLI_BIN_WITH_PATH',
                                       'run_script_input': i['run_script_input'],
                                       'recursion_spaces': recursion_spaces})
    if r['return'] > 0:
        if r['return'] == 16:
            if env.get('CM_TMP_FAIL_IF_NOT_FOUND', '').lower() == 'yes':
                return r

            print(recursion_spaces + '    # {}'.format(r['error']))

            # Attempt to run installer
            r = {
                'return': 0,
                'skip': True,
                'script': {
                    'tags': 'install,github-cli'}}

        return r

    found_path = r['found_path']

    return {'return': 0}


def postprocess(i):
    env = i['env']

    r = i['automation'].parse_version({'match_text': r'gh\s*version\s*([\d.]+)',
                                       'group_number': 1,
                                       'env_key': 'CM_GITHUBCLI_VERSION',
                                       'which_env': i['env']})
    if r['return'] > 0:
        return r

    version = r['version']

    print(i['recursion_spaces'] + '    Detected version: {}'.format(version))

    return {'return': 0, 'version': version}



#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

from cmind import utils
import cmind as cm
import os
import subprocess
from os.path import exists


def preprocess(i):

    os_info = i['os_info']
    env = i['env']
    submission_dir = env.get("CM_MLPERF_INFERENCE_SUBMISSION_DIR", "")

    if submission_dir == "":
        print("Please set CM_MLPERF_INFERENCE_SUBMISSION_DIR")
        return {'return': 1, 'error': 'CM_MLPERF_INFERENCE_SUBMISSION_DIR is not specified in env in run-mlperf-accuracy-log-truncator'}

    submitter = env.get("CM_MLPERF_SUBMITTER", "CTuning")

    os.system("rm -rf " + submission_dir + "_logs")

    CMD = env['CM_PYTHON_BIN'] + " '" + os.path.join(env['CM_MLPERF_INFERENCE_SOURCE'], "tools", "submission",
                                                     "truncate_accuracy_log.py") + "' --input '" + submission_dir + "' --submitter '" + submitter + "' --backup '" + submission_dir + "_logs'"
    env['CM_RUN_CMD'] = CMD

    return {'return': 0}



from cmind import utils
import os
import json


def preprocess(i):

    env = i['env']

    cm_cache_dataset_path = env.get(
        'CM_CUSTOM_CACHE_ENTRY_DATASET_MLCOMMONS_COGNATA_PATH', '').strip()
    cfg = utils.safe_load_json(cm_cache_dataset_path, 'cfg.json')['meta']
    if cfg.get('imported', False):
        env['CM_DATASET_MLCOMMONS_COGNATA_IMPORTED'] = 'yes'

    if env.get('CM_ABTF_SCRATCH_PATH_DATASETS', '') != '':
        env['CM_ABTF_SCRATCH_PATH_DATASET_COGNATA'] = os.path.join(
            env['CM_ABTF_SCRATCH_PATH_DATASETS'], "cognata")
        env['CM_ABTF_SCRATCH_PATH_DATASET_COGNATA_TMP'] = os.path.join(
            env['CM_ABTF_SCRATCH_PATH_DATASETS'], "cognata_tmp")

    env['CM_DATASET_COGNATA_POC_TEXT_MD5_FILE_PATH'] = os.path.join(
        i['run_script_input']['path'], 'checksums', 'cognata_poc.txt')

    # Check if user requests path not in CM cache
    #
    # --path (env CM_TMP_PATH) shows where to store Cognata data set instead of CM cahe
    # --import tells CM to import existing Cognata from a given path and skip further download/processing
    #
    import_path = env.get(
        'CM_DATASET_MLCOMMONS_COGNATA_IMPORT_PATH',
        '').strip()
    if import_path != '':
        if not os.path.isdir(import_path):
            return {'return': 1, 'error': 'directory to import this dataset doesn\'t exist: {}'.format(
                import_path)}

        env['CM_DATASET_MLCOMMONS_COGNATA_IMPORTED'] = 'yes'
        env['CM_DATASET_MLCOMMONS_COGNATA_PATH'] = import_path

    else:
        path = env.get('CM_TMP_PATH', '')
        if path != '':
            env['CM_DATASET_MLCOMMONS_COGNATA_IMPORTED'] = 'no'

            if not os.path.isdir(path):
                os.makedirs(path)

            env['CM_DATASET_MLCOMMONS_COGNATA_PATH'] = path

    return {'return': 0}


def postprocess(i):

    env = i['env']

    automation = i['automation']
    cm = automation.cmind

    cur_dir = os.getcwd()

    quiet = (env.get('CM_QUIET', False) == 'yes')

    cm_cache_dataset_path = env.get(
        'CM_CUSTOM_CACHE_ENTRY_DATASET_MLCOMMONS_COGNATA_PATH', '').strip()

    if not os.path.isdir(cm_cache_dataset_path):
        return {
            'return': 1, 'error': 'Dataset corrupted - CM cache path not found: {}'.format(cm_cache_dataset_path)}

    if env.get('CM_DATASET_MLCOMMONS_COGNATA_FILE_NAMES', '') == '':
        env['CM_DATASET_MLCOMMONS_COGNATA_PATH'] = os.path.dirname(
            env['CM_CUSTOM_CACHE_ENTRY_DATASET_MLCOMMONS_COGNATA_PATH'])
        env['CM_GET_DEPENDENT_CACHED_PATH'] = env['CM_DATASET_MLCOMMONS_COGNATA_PATH']
        return {'return': 0}

    cm_cache_dataset_cfg_file = os.path.join(cm_cache_dataset_path, 'cfg.json')
    env['CM_DATASET_MLCOMMONS_COGNATA_CFG_FILE'] = cm_cache_dataset_cfg_file

    cfg = utils.safe_load_json('', cm_cache_dataset_cfg_file)['meta']

    dataset_path = cfg.get('real_path', '')
    dataset_path_requested = env.get('CM_DATASET_MLCOMMONS_COGNATA_PATH', '')
    if dataset_path == '':
        if dataset_path_requested != '':
            dataset_path = dataset_path_requested
        else:
            dataset_path = os.path.join(cm_cache_dataset_path, 'cognata')
    else:
        if dataset_path_requested != '':
            dataset_path = dataset_path_requested

    cfg['real_path'] = dataset_path

    print('')
    print('Used dataset path: {}'.format(dataset_path))

    env['CM_DATASET_MLCOMMONS_COGNATA_PATH'] = dataset_path

    # If imported, don't process further
    if env.get('CM_DATASET_MLCOMMONS_COGNATA_IMPORTED', '') == 'yes':
        cfg['imported'] = True
    else:
        cfg['imported'] = False

    utils.save_json(cm_cache_dataset_cfg_file, cfg)

    if cfg.get('imported', False):
        return {'return': 0}

    # If processed once, don't process unless forced
    if cfg.get('processed', False):
        if not utils.check_if_true_yes_on(
                env, 'CM_DATASET_MLCOMMONS_COGNATA_UPDATE'):
            print('')
            print('Already processed: use --update to update this dataset')

            return {'return': 0}

    # First level dir
    dataset_path1 = dataset_path

    if not os.path.isdir(dataset_path1):
        os.makedirs(dataset_path1)

    # Check if has license and download URL
    dataset_path_secret = os.path.join(dataset_path1, 'secret.json')

    first_url = ''
    dataset_meta = {}

    if os.path.isfile(dataset_path_secret):
        r = utils.load_json(dataset_path_secret)
        if r['return'] > 0:
            return r

        dataset_meta = r['meta']

        first_url = dataset_meta.get('first_url', '').strip()

    if first_url == '':
        x = env.get('CM_DATASET_MLCOMMONS_COGNATA_PRIVATE_URL', '').strip()
        if x != '':
            first_url = x
        else:
            print('')
            first_url = input(
                'Please register at https://mlcommons.org/datasets/cognata and enter private URL: ')

            first_url = first_url.strip()

        if first_url == '':
            return {'return': 1,
                    'error': 'Private MLCommons Cognata URL was not provided'}

        dataset_meta['first_url'] = first_url

        with open(dataset_path_secret, 'w') as f:
            f.write(json.dumps(dataset_meta, indent=2) + '\n')

    ##########################################################################
    # Check if first.xlsx exists
    file_first_xlsx = 'first.xlsx'
    first_xlsx = os.path.join(dataset_path1, file_first_xlsx)

    if not os.path.isfile(first_xlsx):
        # Attempting to download file
        first_url_export, dummy = google_url_for_export(first_url)

        if first_url_export == '':
            return {
                'return': 1, 'error': 'can\'t parse URL for export: {}'.format(first_url)}

        r = cm.access({'action': 'run',
                       'automation': 'script',
                       'tags': 'download,file,_wget',
                       'verify': 'no',
                       'url': first_url_export,
                       'output_file': file_first_xlsx,
                       'store': dataset_path1})
        if r['return'] > 0:
            return r

    if not os.path.isfile(first_xlsx):
        return {'return': 1,
                'error': 'File {} was not downloaded'.format(first_xlsx)}

    ##########################################################################
    # Parse XLSX and check serial number
    serial_numbers = []
    for s in env.get(
            'CM_DATASET_MLCOMMONS_COGNATA_SERIAL_NUMBERS', '').strip().split(','):
        s = s.strip()
        if s != '' and s not in serial_numbers:
            serial_numbers.append(s)

    dataset_key = env['CM_DATASET_MLCOMMONS_COGNATA_KEY1']
    url_key = 'Link to Excel File (Download Links)'
    serial_key = 'Serial Number'

    r = process_xlsx(
        first_xlsx,
        dataset_key,
        url_key,
        serial_key,
        serial_numbers)
    if r['return'] > 0:
        return r

    headers = r['headers']
    data = r['data']
    all_data = r['all_data']

    if len(all_data) != 0:
        file_first_json = 'first.json'
        first_json = os.path.join(dataset_path1, file_first_json)

        if not os.path.isfile(first_json):
            with open(first_json, 'w') as f:
                f.write(json.dumps(all_data, indent=2) + '\n')

    if len(data) == 0:
        return {'return': 0, 'error': 'no sets found'}

    ##########################################################################
    print('')
    print('Available or selected serial numbers (use --serial_numbers=a,b,c to download specific subsets):')
    print('')
    for d in data:
        s = d[serial_key]
        print(s)

    for d in data:
        url = d[url_key]
        url_export, dummy = google_url_for_export(url)

        serial_file = d[serial_key] + '.xlsx'

        dataset_path2 = os.path.join(dataset_path1, serial_file)
        dataset_path3 = os.path.join(dataset_path1, d[serial_key])

        if not os.path.isdir(dataset_path3):
            os.makedirs(dataset_path3)

        if not os.path.isfile(dataset_path2):

            print('')
            print('Downloading {} ...'.format(url_export))

            r = cm.access({'action': 'run',
                           'automation': 'script',
                           'tags': 'download,file,_wget',
                           'verify': 'no',
                           'url': url_export,
                           'output_file': serial_file,
                           'store': dataset_path1})
            if r['return'] > 0:
                return r

    ##########################################################################
    print('')
    print('Processing subsets ...')

    group_names = []
    for s in env.get('CM_DATASET_MLCOMMONS_COGNATA_GROUP_NAMES',
                     '').strip().split(','):
        s = s.strip()
        if s != '' and s not in group_names:
            group_names.append(s)

    # Check if force some filenames
    x = env.get('CM_DATASET_MLCOMMONS_COGNATA_FILE_NAMES', '').strip()
    file_names = []
    if x != '':
        file_names = x.strip(';') if ';' in x else [x]

    for d in data:
        serial_file = d[serial_key] + '.xlsx'

        dataset_path2 = os.path.join(dataset_path1, serial_file)
        dataset_path3 = os.path.join(dataset_path1, d[serial_key])

        print('')
        print('Processing {} ...'.format(serial_file))

        dataset_key = 'File_Data'
        url_key = 'File_Link'
        serial_key = 'Group_Name'

        r = process_xlsx(
            dataset_path2,
            dataset_key,
            url_key,
            serial_key,
            group_names)
        if r['return'] > 0:
            return r

        headers = r['headers']
        data = r['data']
        all_data = r['all_data']

        if len(all_data) != 0:
            file_all_json = 'all.json'
            all_json = os.path.join(dataset_path3, file_all_json)

            if not os.path.isfile(all_json):
                with open(all_json, 'w') as f:
                    f.write(json.dumps(all_data, indent=2) + '\n')

        if len(data) == 0:
            return {'return': 0, 'error': 'no sub-sets found'}

        for d in data:
            file_name = d['File_Name']

            if len(file_names) > 0 and file_name not in file_names:
                continue

            file_name_with_path = os.path.join(dataset_path3, file_name)
            file_name_with_path_done = os.path.join(
                dataset_path3, file_name) + '.done'

            url = d[url_key]

            print('')
            print('Downloading {} ...'.format(file_name))

            if os.path.isfile(file_name_with_path_done):
                print('')
                print('  Already processed - skipping ...')
                continue

            if os.name == 'nt':
                aria2_tool = env['CM_ARIA2_BIN_WITH_PATH']
            else:
                aria2_tool = 'aria2c'

            cmd = aria2_tool + \
                ' --async-dns=false -x15 -s15 "{}" --dir "{}" -o "{}"'.format(
                    url, dataset_path3, file_name)

            print('')
            print(cmd)
            print('')

            os.system(cmd)

            # Unarchive
            print('')
            print('Extracting file {} ...'.format(file_name_with_path))
            print('')

            if file_name.endswith('.zip'):

                import zipfile
                extractor = zipfile.ZipFile(file_name_with_path, "r")

            elif file_name.endswith('.tar'):

                import tarfile
                extractor = tarfile.ZipFile(file_name_with_path, "r")

            else:
                extractor = None

            if extractor is not None:

                try:
                    extractor.extractall(dataset_path3)
                    extractor.close()

                except Exception as e:
                    return {'return': 1,
                            'error': 'extracting failed: {}'.format(e)}

            # Mark as downloaded
            with open(file_name_with_path_done, 'w') as f:
                f.write('DONE\n')

            # Remove file
            os.remove(file_name_with_path)

    print('')

    # Mark that processed this dataset once correctly
    cfg['processed'] = True
    utils.save_json(cm_cache_dataset_cfg_file, cfg)

    env['CM_GET_DEPENDENT_CACHED_PATH'] = env['CM_DATASET_MLCOMMONS_COGNATA_PATH']

    return {'return': 0}


# Prepare Google URL for export
def google_url_for_export(url):
    url2 = ''

    j = url.rfind('/')

    if j > 0:
        url = url[:j + 1]
        url2 = url + 'export'

    return (url2, url)

# Download Cognata XLSX


def process_xlsx(filename, dataset_key, url_key, serial_key, serial_numbers):
    import openpyxl

    ex = openpyxl.load_workbook(filename)

    sets = ex[dataset_key]

    headers = {}

    data = []
    all_data = []

    for row in sets.iter_rows(values_only=True):
        lrow = list(row)

        if len(headers) == 0:
            for j in range(0, len(lrow)):
                headers[j] = str(lrow[j]).strip()
        else:
            xrow = {}

            for j in range(0, len(lrow)):
                xrow[headers[j]] = lrow[j]

            url = str(xrow.get(url_key, ''))
            if 'https' in url:
                all_data.append(xrow)

                if len(serial_numbers) > 0:
                    serial_number = xrow.get(serial_key, '')

                    if serial_number not in serial_numbers:
                        continue

                if url != '':
                    data.append(xrow)

    return {'return': 0, 'headers': headers,
            'data': data, 'all_data': all_data}



# Author: Grigori Fursin
# Contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#
# Copyright: https://github.com/mlcommons/ck/blob/master/cm-mlops/COPYRIGHT.md
# License: https://github.com/mlcommons/ck/blob/master/cm-mlops/LICENSE.md
#
# White paper: https://arxiv.org/abs/2406.16791
# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
# Original repository: https://github.com/mlcommons/ck/tree/master/cm-mlops
#
# CK and CM project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
#

import os

from cmind.automation import Automation
from cmind import utils


class CAutomation(Automation):
    """
    Automation actions
    """

    ############################################################
    def __init__(self, cmind, automation_file):
        super().__init__(cmind, __file__)

    ############################################################
    def test(self, i):
        """
        Test automation

        Args:
          (CM input dict):

          (out) (str): if 'con', output to console

          automation (str): automation as CM string object

          parsed_automation (list): prepared in CM CLI or CM access function
                                    [ (automation alias, automation UID) ] or
                                    [ (automation alias, automation UID), (automation repo alias, automation repo UID) ]

          (artifact) (str): artifact as CM string object

          (parsed_artifact) (list): prepared in CM CLI or CM access function
                                    [ (artifact alias, artifact UID) ] or
                                    [ (artifact alias, artifact UID), (artifact repo alias, artifact repo UID) ]

          ...

        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

          * Output from this automation action

        """

        import json
        print(json.dumps(i, indent=2))

        return {'return': 0}

    ##########################################################################
    def get_host_os_info(self, i):
        """
        Get some host platform name (currently windows or linux) and OS bits

        Args:
           (CM input dict):

           (bits) (str): force host platform bits

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0

           * info (dict):
             * platform (str): "windows", "linux" or "darwin"
             * bat_ext (str): ".bat" or ".sh"
             * bits (str): 32 or 64 bits
             * python_bits 9str): python bits

        """

        import os
        import platform
        import struct

        info = {}

        pbits = str(8 * struct.calcsize("P"))

        if platform.system().lower().startswith('win'):
            platform = 'windows'
            info['bat_ext'] = '.bat'
            info['set_env'] = 'set ${key}=${value}'
            info['env_separator'] = ';'
            info['env_var'] = '%env_var%'
            info['bat_rem'] = 'rem ${rem}'
            info['run_local_bat'] = 'call ${bat_file}'
            info['run_local_bat_from_python'] = 'call ${bat_file}'
            info['run_bat'] = 'call ${bat_file}'
            info['start_script'] = ['@echo off', '']
            info['env'] = {
                "CM_WINDOWS": "yes"
            }
        else:
            if platform.system().lower().startswith('darwin'):
                platform = 'darwin'
            else:
                platform = 'linux'

            info['bat_ext'] = '.sh'
            info['set_env'] = 'export ${key}="${value}"'
            info['env_separator'] = ':'
            info['env_var'] = '${env_var}'
            info['set_exec_file'] = 'chmod 755 "${file_name}"'
            info['bat_rem'] = '# ${rem}'
            info['run_local_bat'] = '. ./${bat_file}'
            info['run_local_bat_from_python'] = 'bash -c ". ./${bat_file}"'
            info['run_bat'] = '. ${bat_file}'
            info['start_script'] = ['#!/bin/bash', '']
            info['env'] = {}

        info['platform'] = platform

        obits = i.get('bits', '')
        if obits == '':
            obits = '32'
            if platform == 'windows':
                # Trying to get fast way to detect bits
                if os.environ.get('ProgramW6432', '') != '' or os.environ.get(
                        'ProgramFiles(x86)', '') != '':  # pragma: no cover
                    obits = '64'
            else:
                # On Linux use first getconf LONG_BIT and if doesn't work use
                # python bits

                obits = pbits

                r = utils.gen_tmp_file({})
                if r['return'] > 0:
                    return r

                fn = r['file_name']

                cmd = 'getconf LONG_BIT > ' + fn
                rx = os.system(cmd)

                if rx == 0:
                    r = utils.load_txt(file_name=fn, remove_after_read=True)

                    if r['return'] == 0:
                        s = r['string'].strip()
                        if len(s) > 0 and len(s) < 4:
                            obits = s
                else:
                    if os.path.isfile(fn):
                        os.remove(fn)

        info['bits'] = obits
        info['python_bits'] = pbits

        return {'return': 0, 'info': info}

    ##########################################################################
    def download_file(self, i):
        """
        Download file using requests

        Args:
           (CM input dict):

           url (str): URL with file
           (filename) (str): explicit file name
           (path) (str): path to record file (or current if empty)
           (chunk_size) (int): chunck size in bytes (65536 by default)
           (text) (str): print text before downloaded status ("Downloaded: " by default)
           (verify) (bool): verify SSL certificate if True (True by default)
                            can be switched by global env CM_UTILS_DOWNLOAD_VERIFY_SSL = no

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0

           * filename (str): file name
           * path (str): path to file
           * size (int): file size

        """

        import requests
        import time
        import sys
        from urllib import parse

        # Get URL
        url = i['url']

        # Check file name
        file_name = i.get('filename', '')
        if file_name == '':
            parsed_url = parse.urlparse(url)
            file_name = os.path.basename(parsed_url.path)

        # Check path
        path = i.get('path', '')
        if path is None or path == '':
            path = os.getcwd()

        # Output file
        path_to_file = os.path.join(path, file_name)

        if os.path.isfile(path_to_file):
            os.remove(path_to_file)

        print('Downloading to {}'.format(path_to_file))
        print('')

        # Download
        size = -1
        downloaded = 0
        chunk_size = i.get('chunk_size', 65536)

        text = i.get('text', 'Downloaded: ')

        if 'CM_UTILS_DOWNLOAD_VERIFY_SSL' in os.environ:
            verify = os.environ['CM_UTILS_DOWNLOAD_VERIFY_SSL'] == 'yes'
        else:
            verify = i.get('verify', True)

        try:
            with requests.get(url, stream=True, allow_redirects=True, verify=verify) as download:
                download.raise_for_status()

                size_string = download.headers.get('Content-Length')

                if size_string is None:
                    transfer_encoding = download.headers.get(
                        'Transfer-Encoding', '')
                    if transfer_encoding != 'chunked':
                        return {'return': 1, 'error': 'did not receive file'}
                    else:
                        size_string = "0"

                size = int(size_string)

                with open(path_to_file, 'wb') as output:
                    for chunk in download.iter_content(chunk_size=chunk_size):

                        if chunk:
                            output.write(chunk)
                        if size == 0:
                            continue
                        downloaded += 1
                        percent = downloaded * chunk_size * 100 / size

                        sys.stdout.write("\r{}{:3.0f}%".format(text, percent))
                        sys.stdout.flush()

                    sys.stdout.write("\r{}{:3.0f}%".format(text, 100))
                    sys.stdout.flush()

        except Exception as e:
            return {'return': 1, 'error': format(e)}

        print('')
        if size == 0:
            file_stats = os.stat(path_to_file)
            size = file_stats.st_size

        return {'return': 0, 'filename': file_name,
                'path': path_to_file, 'size': size}

    ##########################################################################
    def unzip_file(self, i):
        """
        Unzip file

        Args:
           (CM input dict):

           filename (str): explicit file name
           (path) (str): path where to unzip file (current path otherwise)
           (strip_folders) (int): strip first folders

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0

        """

        import zipfile

        # Check file name
        file_name = i['filename']

        if not os.path.isfile(file_name):
            return {'return': 1,
                    'error': 'file {} not found'.format(file_name)}

        console = i.get('out') == 'con'

        # Attempt to read cmr.json
        file_name_handle = open(file_name, 'rb')
        file_name_zip = zipfile.ZipFile(file_name_handle)

        info_files = file_name_zip.infolist()

        path = i.get('path', '')
        if path is None or path == '':
            path = os.getcwd()

        strip_folders = i.get('strip_folders', 0)

        # Unpacking zip
        for info in info_files:
            f = info.filename
            permissions = info.external_attr

            if not f.startswith('..') and not f.startswith(
                    '/') and not f.startswith('\\'):
                f_zip = f

                if strip_folders > 0:
                    fsplit = f.split('/')  # Zip standard on all OS
                    f = '/'.join(fsplit[strip_folders:])

                file_path = os.path.join(path, f)

                if f.endswith('/'):
                    # create directory
                    if not os.path.exists(file_path):
                        os.makedirs(file_path)
                else:
                    dir_name = os.path.dirname(file_path)
                    if not os.path.exists(dir_name):
                        os.makedirs(dir_name)

                    # extract file
                    file_out = open(file_path, 'wb')
                    file_out.write(file_name_zip.read(f_zip))
                    file_out.close()

                    if permissions > 0xffff:
                        os.chmod(file_path, permissions >> 16)

        file_name_zip.close()
        file_name_handle.close()

        return {'return': 0}

    ##########################################################################
    def compare_versions(self, i):
        """
        Compare versions

        Args:

           version1 (str): version 1
           version2 (str): version 2

        Returns:
           (CM return dict):

           * comparison (int):  1 - version 1 > version 2
                                0 - version 1 == version 2
                               -1 - version 1 < version 2

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        version1 = i['version1']
        version2 = i['version2']

        l_version1 = version1.split('.')
        l_version2 = version2.split('.')

        # 3.9.6 vs 3.9
        # 3.9 vs 3.9.6

        i_version1 = [int(v) if v.isdigit() else v for v in l_version1]
        i_version2 = [int(v) if v.isdigit() else v for v in l_version2]

        comparison = 0

        for index in range(max(len(i_version1), len(i_version2))):
            v1 = i_version1[index] if index < len(i_version1) else 0
            v2 = i_version2[index] if index < len(i_version2) else 0

            if v1 > v2:
                comparison = 1
                break
            elif v1 < v2:
                comparison = -1
                break

        return {'return': 0, 'comparison': comparison}

    ##########################################################################
    def json2yaml(self, i):
        """
        Convert JSON file to YAML

        Args:

           input (str): input file (.json)
           (output) (str): output file (.yaml)

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        input_file = i.get('input', '')

        if input_file == '':
            return {'return': 1, 'error': 'please specify --input={json file}'}

        output_file = i.get('output', '')

        r = utils.load_json(input_file, check_if_exists=True)
        if r['return'] > 0:
            return r

        meta = r['meta']

        if output_file == '':
            output_file = input_file[:-
                                     5] if input_file.endswith('.json') else input_file
            output_file += '.yaml'

        r = utils.save_yaml(output_file, meta)
        if r['return'] > 0:
            return r

        return {'return': 0}

    ##########################################################################
    def yaml2json(self, i):
        """
        Convert YAML file to JSON

        Args:

           input (str): input file (.yaml)
           (output) (str): output file (.json)

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        input_file = i.get('input', '')

        if input_file == '':
            return {'return': 1, 'error': 'please specify --input={yaml file}'}

        output_file = i.get('output', '')

        r = utils.load_yaml(input_file, check_if_exists=True)
        if r['return'] > 0:
            return r

        meta = r['meta']

        if output_file == '':
            output_file = input_file[:-
                                     5] if input_file.endswith('.yaml') else input_file
            output_file += '.json'

        r = utils.save_json(output_file, meta)
        if r['return'] > 0:
            return r

        return {'return': 0}

    ##########################################################################
    def sort_json(self, i):
        """
        Sort JSON file

        Args:

           input (str): input file (.json)
           (output) (str): output file

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        input_file = i.get('input', '')

        if input_file == '':
            return {'return': 1, 'error': 'please specify --input={json file}'}

        r = utils.load_json(input_file, check_if_exists=True)
        if r['return'] > 0:
            return r

        meta = r['meta']

        output_file = i.get('output', '')

        if output_file == '':
            output_file = input_file

        r = utils.save_json(output_file, meta, sort_keys=True)
        if r['return'] > 0:
            return r

        return {'return': 0}

    ##########################################################################
    def dos2unix(self, i):
        """
        Convert DOS file to UNIX (remove \r)

        Args:

           input (str): input file (.txt)
           (output) (str): output file

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        input_file = i.get('input', '')

        if input_file == '':
            return {'return': 1, 'error': 'please specify --input={txt file}'}

        r = utils.load_txt(input_file, check_if_exists=True)
        if r['return'] > 0:
            return r

        s = r['string'].replace('\r', '')

        output_file = i.get('output', '')

        if output_file == '':
            output_file = input_file

        r = utils.save_txt(output_file, s)
        if r['return'] > 0:
            return r

        return {'return': 0}

    ##########################################################################
    def replace_string_in_file(self, i):
        """
        Convert DOS file to UNIX (remove \r)

        Args:

           input (str): input file (.txt)
           (output) (str): output file
           string (str): string to replace
           replacement (str): replacement string

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0

           (update) (bool): True if file was upated
        """

        input_file = i.get('input', '')
        if input_file == '':
            return {'return': 1, 'error': 'please specify --input={txt file}'}

        string = i.get('string', '')
        if string == '':
            return {'return': 1,
                    'error': 'please specify --string={string to replace}'}

        replacement = i.get('replacement', '')
        if replacement == '':
            return {'return': 1,
                    'error': 'please specify --replacement={string to replace}'}

        output_file = i.get('output', '')

        if output_file == '':
            output_file = input_file

        r = utils.load_txt(input_file, check_if_exists=True)
        if r['return'] > 0:
            return r

        s = r['string'].replace('\r', '')

        s = s.replace(string, replacement)

        r = utils.save_txt(output_file, s)
        if r['return'] > 0:
            return r

        return {'return': 0}

    ##########################################################################
    def create_toc_from_md(self, i):
        """
        Convert DOS file to UNIX (remove \r)

        Args:

           input (str): input file (.md)
           (output) (str): output file (input+'.toc)

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        input_file = i.get('input', '')
        if input_file == '':
            return {'return': 1, 'error': 'please specify --input={txt file}'}

        output_file = i.get('output', '')

        if output_file == '':
            output_file = input_file + '.toc'

        r = utils.load_txt(input_file, check_if_exists=True)
        if r['return'] > 0:
            return r

        lines = r['string'].split('\n')

        toc = []

        toc.append('<details>')
        toc.append('<summary>Click here to see the table of contents.</summary>')
        toc.append('')

        for line in lines:
            line = line.strip()

            if line.startswith('#'):
                j = line.find(' ')
                if j >= 0:
                    title = line[j:].strip()

                    x = title.lower().replace(' ', '-')

                    for k in range(0, 2):
                        if x.startswith('*'):
                            x = x[1:]
                        if x.endswith('*'):
                            x = x[:-1]

                    for z in [':', '+', '.', '(', ')', ',']:
                        x = x.replace(z, '')

                    y = ' ' * (2 * (j - 1)) + '* [' + title + '](#' + x + ')'

                    toc.append(y)

        toc.append('')
        toc.append('</details>')

        r = utils.save_txt(output_file, '\n'.join(toc) + '\n')
        if r['return'] > 0:
            return r

        return {'return': 0}

    ##########################################################################
    def copy_to_clipboard(self, i):
        """
        Copy string to a clipboard

        Args:

           string (str): string to copy to a clipboard
           (add_quotes) (bool): add quotes to the string in a clipboard
           (skip_fail) (bool): if True, do not fail

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        s = i.get('string', '')

        if i.get('add_quotes', False):
            s = '"' + s + '"'

        failed = False
        warning = ''

        # Try to load pyperclip (seems to work fine on Windows)
        try:
            import pyperclip
        except Exception as e:
            warning = format(e)
            failed = True
            pass

        if not failed:
            pyperclip.copy(s)
        else:
            failed = False

            # Try to load Tkinter
            try:
                from Tkinter import Tk
            except ImportError as e:
                warning = format(e)
                failed = True
                pass

            if failed:
                failed = False
                try:
                    from tkinter import Tk
                except ImportError as e:
                    warning = format(e)
                    failed = True
                    pass

            if not failed:
                # Copy to clipboard
                try:
                    r = Tk()
                    r.withdraw()
                    r.clipboard_clear()
                    r.clipboard_append(s)
                    r.update()
                    r.destroy()
                except Exception as e:
                    failed = True
                    warning = format(e)

        rr = {'return': 0}

        if failed:
            if not i.get('skip_fail', False):
                return {'return': 1, 'error': warning}

            rr['warning'] = warning

        return rr

    ##########################################################################
    def list_files_recursively(self, i):
        """
        List files and concatenate into string separate by comma

        Args:

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        files = os.walk('.')

        s = ''

        for (dir_path, dir_names, file_names) in files:
            for f in file_names:
                if s != '':
                    s += ','

                if dir_path == '.':
                    dir_path2 = ''
                else:
                    dir_path2 = dir_path[2:].replace('\\', '/') + '/'

                s += dir_path2 + f

        print(s)

        return {'return': 0}

    ##########################################################################
    def generate_secret(self, i):
        """
        Generate secret for web apps

        Args:

        Returns:
           (CM return dict):

           secret (str): secret

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        import secrets
        s = secrets.token_urlsafe(16)

        print(s)

        return {'return': 0, 'secret': s}

    ##########################################################################
    def detect_tags_in_artifact(self, i):
        """
        Detect if there are tags in an artifact name (spaces) and update input

        Args:

           input (dict) : original input

        Returns:
           (CM return dict):

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        inp = i['input']

        artifact = inp.get('artifact', '')
        if artifact == '.':
            del (inp['artifact'])
        elif ' ' in artifact:  # or ',' in artifact:
            del (inp['artifact'])
            if 'parsed_artifact' in inp:
                del (inp['parsed_artifact'])
            # Force substitute tags
            inp['tags'] = artifact.replace(' ', ',')

        return {'return': 0}

    ##########################################################################
    def prune_input(self, i):
        """
        Leave only input keys and remove the rest (to regenerate CM commands)

        Args:

           input (dict) : original input
           (extra_keys_starts_with) (list): remove keys that starts
                                            with the ones from this list

        Returns:
           (CM return dict):

           new_input (dict): pruned input

           * return (int): return code == 0 if no error and >0 if error
           * (error) (str): error string if return>0
        """

        import copy

        inp = i['input']
        extra_keys = i.get('extra_keys_starts_with', [])

        i_run_cmd_arc = copy.deepcopy(inp)
        for k in inp:
            remove = False
            if k in ['action', 'automation', 'cmd', 'out',
                     'parsed_automation', 'parsed_artifact', 'self_module']:
                remove = True
            if not remove:
                for ek in extra_keys:
                    if k.startswith(ek):
                        remove = True
                        break

            if remove:
                del (i_run_cmd_arc[k])

        return {'return': 0, 'new_input': i_run_cmd_arc}

    ##########################################################################

    def uid(self, i):
        """
        Generate CM UID.

        Args:
          (CM input dict): empty dict

        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

          * uid (str): CM UID
        """

        console = i.get('out') == 'con'

        r = utils.gen_uid()

        if console:
            print(r['uid'])

        return r

    ##########################################################################

    def system(self, i):
        """
        Run system command and redirect output to string.

        Args:
          (CM input dict):

          * cmd (str): command line
          * (path) (str): go to this directory and return back to current
          * (stdout) (str): stdout file
          * (stderr) (str): stderr file

        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

          * ret (int): return code
          * std (str): stdout + stderr
          * stdout (str): stdout
          * stderr (str): stderr
        """

        cmd = i['cmd']

        if cmd == '':
            return {'return': 1, 'error': 'cmd is empty'}

        path = i.get('path', '')
        if path != '' and os.path.isdir(path):
            cur_dir = os.getcwd()
            os.chdir(path)

        if i.get('stdout', '') != '':
            fn1 = i['stdout']
            fn1_delete = False
        else:
            r = utils.gen_tmp_file({})
            if r['return'] > 0:
                return r
            fn1 = r['file_name']
            fn1_delete = True

        if i.get('stderr', '') != '':
            fn2 = i['stderr']
            fn2_delete = False
        else:
            r = utils.gen_tmp_file({})
            if r['return'] > 0:
                return r
            fn2 = r['file_name']
            fn2_delete = True

        cmd += ' > ' + fn1 + ' 2> ' + fn2
        rx = os.system(cmd)

        std = ''
        stdout = ''
        stderr = ''

        if os.path.isfile(fn1):
            r = utils.load_txt(file_name=fn1, remove_after_read=fn1_delete)
            if r['return'] == 0:
                stdout = r['string'].strip()

        if os.path.isfile(fn2):
            r = utils.load_txt(file_name=fn2, remove_after_read=fn2_delete)
            if r['return'] == 0:
                stderr = r['string'].strip()

        std = stdout
        if stderr != '':
            if std != '':
                std += '\n'
            std += stderr

        if path != '' and os.path.isdir(path):
            os.chdir(cur_dir)

        return {'return': 0, 'ret': rx, 'stdout': stdout,
                'stderr': stderr, 'std': std}

    ############################################################
    def load_cfg(self, i):
        """
        Load configuration artifacts and files

        Args:
          (CM input dict):


        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

        """

        return utils.call_internal_module(
            self, __file__, 'module_cfg', 'load_cfg', i)

    ############################################################
    def select_cfg(self, i):
        """
        Select cfg interactively

        Args:
          (CM input dict):
            tags (str): list of tags to find cfg
            alias (str): alias of a cfg file

        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

        """

        i['self_module'] = self

        return utils.call_internal_module(
            self, __file__, 'module_cfg', 'select_cfg', i)

    ############################################################
    def print_yaml(self, i):
        """
        Print YAML file

        Args:
          (CM input dict):
            file (str): input file

        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

        """

        filename = i.get('file', '')
        if filename == '':
            return {'return': 1, 'error': 'please specify --file={YAML file}'}

        r = utils.load_yaml(filename, check_if_exists=True)
        if r['return'] > 0:
            return r

        meta = r['meta']

        import json
        print(json.dumps(meta, indent=2))

        return {'return': 0}

    ############################################################
    def print_json(self, i):
        """
        Print YAML file

        Args:
          (CM input dict):
            file (str): input file

        Returns:
          (CM return dict):

          * return (int): return code == 0 if no error and >0 if error
          * (error) (str): error string if return>0

        """

        filename = i.get('file', '')
        if filename == '':
            return {'return': 1, 'error': 'please specify --file={JSON file}'}

        r = utils.load_json(filename, check_if_exists=True)
        if r['return'] > 0:
            return r

        meta = r['meta']

        import json
        print(json.dumps(meta, indent=2))

        return {'return': 0}
